@@ -1,13 +1,19 @@
-# pylint:disable=too-many-public-methods
 """global configuration / base class for pydantic models used to make simulation."""
 from __future__ import annotations
 
 import json
+import pathlib
+import os
+import tempfile
 from functools import wraps
+from typing import List, Callable, Dict, Union, Tuple, Any
+from math import ceil
+import io
+import hashlib
 
 import rich
-import pydantic
-from pydantic.fields import ModelField
+import pydantic.v1 as pydantic
+from pydantic.v1.fields import ModelField
 import yaml
 import numpy as np
 import h5py
@@ -15,12 +21,16 @@ import xarray as xr
 
 from .types import ComplexNumber, Literal, TYPE_TAG_STR
 from .data.data_array import DataArray, DATA_ARRAY_MAP
+from .file_util import compress_file_to_gzip, extract_gzip_file
 from ..exceptions import FileError
 from ..log import log
 
-# default indentation (# spaces) in files
-INDENT = 4
+
+INDENT_JSON_FILE = 4  # default indentation of json string in json files
+INDENT = None  # default indentation of json string used internally
 JSON_TAG = "JSON_STRING"
+# If json string is larger than ``MAX_STRING_LENGTH``, split the string when storing in hdf5
+MAX_STRING_LENGTH = 1e9
 
 
 def cache(prop):
@@ -33,13 +43,13 @@ def cache(prop):
     def cached_property_getter(self):
         """The new property method to be returned by decorator."""
 
-        stored_value = self._cached_properties.get(prop_name)  # pylint:disable=protected-access
+        stored_value = self._cached_properties.get(prop_name)
 
         if stored_value is not None:
             return stored_value
 
         computed_value = prop(self)
-        self._cached_properties[prop_name] = computed_value  # pylint:disable=protected-access
+        self._cached_properties[prop_name] = computed_value
         return computed_value
 
     return cached_property_getter
@@ -58,18 +68,73 @@ def ndarray_encoder(val):
     return val.real.tolist()
 
 
+def _get_valid_extension(fname: str) -> str:
+    """Return the file extension from fname, validated to accepted ones."""
+    valid_extensions = [".json", ".yaml", ".hdf5", ".h5", ".hdf5.gz"]
+    extensions = [s.lower() for s in pathlib.Path(fname).suffixes[-2:]]
+    if len(extensions) == 0:
+        raise FileError(f"File '{fname}' missing extension.")
+    single_extension = extensions[-1]
+    if single_extension in valid_extensions:
+        return single_extension
+    double_extension = "".join(extensions)
+    if double_extension in valid_extensions:
+        return double_extension
+    raise FileError(
+        f"File extension must be one of {', '.join(valid_extensions)}; file '{fname}' does not "
+        "match any of those."
+    )
+
+
+def skip_if_fields_missing(fields: List[str]):
+    """Decorate ``validator`` to check that other fields have passed validation."""
+
+    def actual_decorator(validator):
+        @wraps(validator)
+        def _validator(cls, val, values):
+            """New validator function."""
+            for field in fields:
+                if field not in values:
+                    log.warning(
+                        f"Could not execute validator '{validator.__name__}' because field "
+                        f"'{field}' failed validation."
+                    )
+                    return val
+
+            return validator(cls, val, values)
+
+        return _validator
+
+    return actual_decorator
+
+
 class Tidy3dBaseModel(pydantic.BaseModel):
     """Base pydantic model that all Tidy3d components inherit from.
     Defines configuration for handling data structures
-    as well as methods for imporing, exporting, and hashing tidy3d objects.
+    as well as methods for importing, exporting, and hashing tidy3d objects.
     For more details on pydantic base models, see:
     `Pydantic Models <https://pydantic-docs.helpmanual.io/usage/models/>`_
     """
 
+    def __hash__(self) -> int:
+        """Hash method."""
+        try:
+            return super().__hash__(self)
+        except TypeError:
+            return hash(self.json())
+
+    def _hash_self(self) -> str:
+        """Hash this component with ``hashlib`` in a way that is the same every session."""
+        bf = io.BytesIO()
+        self.to_hdf5(bf)
+        return hashlib.sha256(bf.getvalue()).hexdigest()
+
     def __init__(self, **kwargs):
         """Init method, includes post-init validators."""
+        log.begin_capture()
         super().__init__(**kwargs)
         self._post_init_validators()
+        log.end_capture(self)
 
     def _post_init_validators(self) -> None:
         """Call validators taking ``self`` that get run after init, implement in subclasses."""
@@ -80,7 +145,7 @@ class Tidy3dBaseModel(pydantic.BaseModel):
         cls.add_type_field()
         cls.generate_docstring()
 
-    class Config:  # pylint: disable=too-few-public-methods
+    class Config:
         """Sets config for all :class:`Tidy3dBaseModel` objects.
 
         Configuration Options
@@ -107,7 +172,7 @@ class Tidy3dBaseModel(pydantic.BaseModel):
         json_encoders = {
             np.ndarray: ndarray_encoder,
             complex: lambda x: ComplexNumber(real=x.real, imag=x.imag),
-            xr.DataArray: DataArray._json_encoder,  # pylint:disable=unhashable-member, protected-access
+            xr.DataArray: DataArray._json_encoder,
         }
         frozen = True
         allow_mutation = False
@@ -143,14 +208,14 @@ class Tidy3dBaseModel(pydantic.BaseModel):
 
     @classmethod
     def from_file(cls, fname: str, group_path: str = None, **parse_obj_kwargs) -> Tidy3dBaseModel:
-        """Loads a :class:`Tidy3dBaseModel` from .yaml, .json, or .hdf5 file.
+        """Loads a :class:`Tidy3dBaseModel` from .yaml, .json, .hdf5, or .hdf5.gz file.
 
         Parameters
         ----------
         fname : str
-            Full path to the .yaml or .json file to load the :class:`Tidy3dBaseModel` from.
+            Full path to the file to load the :class:`Tidy3dBaseModel` from.
         group_path : str, optional
-            Path to a group inside the file to use as the base level. Only for ``.hdf5`` files.
+            Path to a group inside the file to use as the base level. Only for hdf5 files.
             Starting `/` is optional.
         **parse_obj_kwargs
             Keyword arguments passed to either pydantic's ``parse_obj`` function when loading model.
@@ -158,7 +223,7 @@ class Tidy3dBaseModel(pydantic.BaseModel):
         Returns
         -------
         :class:`Tidy3dBaseModel`
-            An instance of the component class calling `load`.
+            An instance of the component class calling ``load``.
 
         Example
         -------
@@ -169,12 +234,12 @@ class Tidy3dBaseModel(pydantic.BaseModel):
 
     @classmethod
     def dict_from_file(cls, fname: str, group_path: str = None) -> dict:
-        """Loads a dictionary containing the model from a .yaml, .json, or .hdf5 file.
+        """Loads a dictionary containing the model from a .yaml, .json, .hdf5, or .hdf5.gz file.
 
         Parameters
         ----------
         fname : str
-            Full path to the .yaml or .json file to load the :class:`Tidy3dBaseModel` from.
+            Full path to the file to load the :class:`Tidy3dBaseModel` from.
         group_path : str, optional
             Path to a group inside the file to use as the base level.
 
@@ -188,19 +253,23 @@ class Tidy3dBaseModel(pydantic.BaseModel):
         >>> simulation = Simulation.from_file(fname='folder/sim.json') # doctest: +SKIP
         """
 
-        if (".json" in fname or ".yaml" in fname) and (group_path is not None):
-            log.warning("'group_path' provided, but this feature only works with '.hdf5' files.")
+        extension = _get_valid_extension(fname)
+        kwargs = {"fname": fname}
 
-        if ".json" in fname:
-            return cls.dict_from_json(fname=fname)
-        if ".yaml" in fname:
-            return cls.dict_from_yaml(fname=fname)
-        if ".hdf5" in fname:
-            if group_path is None:
-                return cls.dict_from_hdf5(fname=fname)
-            return cls.dict_from_hdf5(fname=fname, group_path=group_path)
+        if group_path is not None:
+            if extension == ".hdf5" or extension == ".hdf5.gz":
+                kwargs["group_path"] = group_path
+            else:
+                log.warning("'group_path' provided, but this feature only works with hdf5 files.")
 
-        raise FileError(f"File must be .json, .yaml, or .hdf5 type, given {fname}")
+        converter = {
+            ".json": cls.dict_from_json,
+            ".yaml": cls.dict_from_yaml,
+            ".hdf5": cls.dict_from_hdf5,
+            ".hdf5.gz": cls.dict_from_hdf5_gz,
+            ".h5": cls.dict_from_hdf5,
+        }[extension]
+        return converter(**kwargs)
 
     def to_file(self, fname: str) -> None:
         """Exports :class:`Tidy3dBaseModel` instance to .yaml, .json, or .hdf5 file
@@ -215,14 +284,14 @@ class Tidy3dBaseModel(pydantic.BaseModel):
         >>> simulation.to_file(fname='folder/sim.json') # doctest: +SKIP
         """
 
-        if ".json" in fname:
-            return self.to_json(fname=fname)
-        if ".yaml" in fname:
-            return self.to_yaml(fname=fname)
-        if ".hdf5" in fname:
-            return self.to_hdf5(fname=fname)
-
-        raise FileError(f"File must be .json, .yaml, or .hdf5 type, given {fname}")
+        extension = _get_valid_extension(fname)
+        converter = {
+            ".json": self.to_json,
+            ".yaml": self.to_yaml,
+            ".hdf5": self.to_hdf5,
+            ".hdf5.gz": self.to_hdf5_gz,
+        }[extension]
+        return converter(fname=fname)
 
     @classmethod
     def from_json(cls, fname: str, **parse_obj_kwargs) -> Tidy3dBaseModel:
@@ -265,7 +334,7 @@ class Tidy3dBaseModel(pydantic.BaseModel):
         -------
         >>> sim_dict = Simulation.dict_from_json(fname='folder/sim.json') # doctest: +SKIP
         """
-        with open(fname, "r", encoding="utf-8") as json_fhandle:
+        with open(fname, encoding="utf-8") as json_fhandle:
             model_dict = json.load(json_fhandle)
         return model_dict
 
@@ -281,7 +350,7 @@ class Tidy3dBaseModel(pydantic.BaseModel):
         -------
         >>> simulation.to_json(fname='folder/sim.json') # doctest: +SKIP
         """
-        json_string = self._json_string
+        json_string = self._json(indent=INDENT_JSON_FILE)
         self._warn_if_contains_data(json_string)
         with open(fname, "w", encoding="utf-8") as file_handle:
             file_handle.write(json_string)
@@ -327,7 +396,7 @@ class Tidy3dBaseModel(pydantic.BaseModel):
         -------
         >>> sim_dict = Simulation.dict_from_yaml(fname='folder/sim.yaml') # doctest: +SKIP
         """
-        with open(fname, "r", encoding="utf-8") as yaml_in:
+        with open(fname, encoding="utf-8") as yaml_in:
             model_dict = yaml.safe_load(yaml_in)
         return model_dict
 
@@ -347,7 +416,7 @@ class Tidy3dBaseModel(pydantic.BaseModel):
         self._warn_if_contains_data(json_string)
         model_dict = json.loads(json_string)
         with open(fname, "w+", encoding="utf-8") as file_handle:
-            yaml.dump(model_dict, file_handle, indent=INDENT)
+            yaml.dump(model_dict, file_handle, indent=INDENT_JSON_FILE)
 
     @staticmethod
     def _warn_if_contains_data(json_str: str) -> None:
@@ -402,8 +471,27 @@ class Tidy3dBaseModel(pydantic.BaseModel):
                     model_dict = model_dict[key]
         return model_dict
 
+    @staticmethod
+    def _json_string_key(index: int) -> str:
+        """Get json string key for string chunk number ``index``."""
+        if index:
+            return f"{JSON_TAG}_{index}"
+        return JSON_TAG
+
     @classmethod
-    def dict_from_hdf5(cls, fname: str, group_path: str = "") -> dict:
+    def _json_string_from_hdf5(cls, fname: str) -> str:
+        """Load the model json string from an hdf5 file."""
+        with h5py.File(fname, "r") as f_handle:
+            num_string_parts = len([key for key in f_handle.keys() if JSON_TAG in key])
+            json_string = b""
+            for ind in range(num_string_parts):
+                json_string += f_handle[cls._json_string_key(ind)][()]
+        return json_string
+
+    @classmethod
+    def dict_from_hdf5(
+        cls, fname: str, group_path: str = "", custom_decoders: List[Callable] = None
+    ) -> dict:
         """Loads a dictionary containing the model contents from a .hdf5 file.
 
         Parameters
@@ -412,6 +500,10 @@ class Tidy3dBaseModel(pydantic.BaseModel):
             Full path to the .hdf5 file to load the :class:`Tidy3dBaseModel` from.
         group_path : str, optional
             Path to a group inside the file to selectively load a sub-element of the model only.
+        custom_decoders : List[Callable]
+            List of functions accepting
+            (fname: str, group_path: str, model_dict: dict, key: str, value: Any) that store the
+            value in the model dict after a custom decoding.
 
         Returns
         -------
@@ -423,15 +515,29 @@ class Tidy3dBaseModel(pydantic.BaseModel):
         >>> sim_dict = Simulation.dict_from_hdf5(fname='folder/sim.hdf5') # doctest: +SKIP
         """
 
+        def is_data_array(value: Any) -> bool:
+            """Whether a value is supposed to be a data array based on the contents."""
+            return isinstance(value, str) and value in DATA_ARRAY_MAP
+
         def load_data_from_file(model_dict: dict, group_path: str = "") -> None:
             """For every DataArray item in dictionary, load path of hdf5 group as value."""
 
             for key, value in model_dict.items():
-
                 subpath = f"{group_path}/{key}"
 
+                # apply custom validation to the key value pair and modify model_dict
+                if custom_decoders:
+                    for custom_decoder in custom_decoders:
+                        custom_decoder(
+                            fname=fname,
+                            group_path=subpath,
+                            model_dict=model_dict,
+                            key=key,
+                            value=value,
+                        )
+
                 # write the path to the element of the json dict where the data_array should be
-                if isinstance(value, str) and value in DATA_ARRAY_MAP:
+                if is_data_array(value):
                     data_array_type = DATA_ARRAY_MAP[value]
                     model_dict[key] = data_array_type.from_hdf5(fname=fname, group_path=subpath)
                     continue
@@ -441,21 +547,30 @@ class Tidy3dBaseModel(pydantic.BaseModel):
                     value_dict = cls.tuple_to_dict(tuple_values=value)
                     load_data_from_file(model_dict=value_dict, group_path=subpath)
 
+                    # handle case of nested list of DataArray elements
+                    val_tuple = list(value_dict.values())
+                    for ind, (model_item, value_item) in enumerate(zip(model_dict[key], val_tuple)):
+                        if is_data_array(model_item):
+                            model_dict[key][ind] = value_item
+
                 # if a dict, recurse
                 elif isinstance(value, dict):
                     load_data_from_file(model_dict=value, group_path=subpath)
 
-        with h5py.File(fname, "r") as f_handle:
-            json_string = f_handle[JSON_TAG][()]
-            model_dict = json.loads(json_string)
-
+        model_dict = json.loads(cls._json_string_from_hdf5(fname=fname))
         group_path = cls._construct_group_path(group_path)
         model_dict = cls.get_sub_model(group_path=group_path, model_dict=model_dict)
         load_data_from_file(model_dict=model_dict, group_path=group_path)
         return model_dict
 
     @classmethod
-    def from_hdf5(cls, fname: str, group_path: str = "", **parse_obj_kwargs) -> Tidy3dBaseModel:
+    def from_hdf5(
+        cls,
+        fname: str,
+        group_path: str = "",
+        custom_decoders: List[Callable] = None,
+        **parse_obj_kwargs,
+    ) -> Tidy3dBaseModel:
         """Loads :class:`Tidy3dBaseModel` instance to .hdf5 file.
 
         Parameters
@@ -465,25 +580,34 @@ class Tidy3dBaseModel(pydantic.BaseModel):
         group_path : str, optional
             Path to a group inside the file to selectively load a sub-element of the model only.
             Starting `/` is optional.
+        custom_decoders : List[Callable]
+            List of functions accepting
+            (fname: str, group_path: str, model_dict: dict, key: str, value: Any) that store the
+            value in the model dict after a custom decoding.
         **parse_obj_kwargs
             Keyword arguments passed to pydantic's ``parse_obj`` method.
 
         Example
         -------
-        >>> simulation.to_hdf5(fname='folder/sim.hdf5') # doctest: +SKIP
+        >>> simulation = Simulation.from_hdf5(fname='folder/sim.hdf5') # doctest: +SKIP
         """
 
         group_path = cls._construct_group_path(group_path)
-        model_dict = cls.dict_from_hdf5(fname=fname, group_path=group_path)
+        model_dict = cls.dict_from_hdf5(
+            fname=fname, group_path=group_path, custom_decoders=custom_decoders
+        )
         return cls.parse_obj(model_dict, **parse_obj_kwargs)
 
-    def to_hdf5(self, fname: str) -> None:
+    def to_hdf5(self, fname: str, custom_encoders: List[Callable] = None) -> None:
         """Exports :class:`Tidy3dBaseModel` instance to .hdf5 file.
 
         Parameters
         ----------
         fname : str
             Full path to the .hdf5 file to save the :class:`Tidy3dBaseModel` to.
+        custom_encoders : List[Callable]
+            List of functions accepting (fname: str, group_path: str, value: Any) that take
+            the ``value`` supplied and write it to the hdf5 ``fname`` at ``group_path``.
 
         Example
         -------
@@ -491,17 +615,23 @@ class Tidy3dBaseModel(pydantic.BaseModel):
         """
 
         with h5py.File(fname, "w") as f_handle:
-
-            f_handle[JSON_TAG] = self._json_string
+            json_str = self._json_string
+            for ind in range(ceil(len(json_str) / MAX_STRING_LENGTH)):
+                ind_start = int(ind * MAX_STRING_LENGTH)
+                ind_stop = min(int(ind + 1) * MAX_STRING_LENGTH, len(json_str))
+                f_handle[self._json_string_key(ind)] = json_str[ind_start:ind_stop]
 
             def add_data_to_file(data_dict: dict, group_path: str = "") -> None:
                 """For every DataArray item in dictionary, write path of hdf5 group as value."""
 
                 for key, value in data_dict.items():
-
                     # append the key to the path
                     subpath = f"{group_path}/{key}"
 
+                    if custom_encoders:
+                        for custom_encoder in custom_encoders:
+                            custom_encoder(fname=f_handle, group_path=subpath, value=value)
+
                     # write the path to the element of the json dict where the data_array should be
                     if isinstance(value, xr.DataArray):
                         value.to_hdf5(fname=f_handle, group_path=subpath)
@@ -517,6 +647,103 @@ class Tidy3dBaseModel(pydantic.BaseModel):
 
             add_data_to_file(data_dict=self.dict())
 
+    @classmethod
+    def dict_from_hdf5_gz(
+        cls, fname: str, group_path: str = "", custom_decoders: List[Callable] = None
+    ) -> dict:
+        """Loads a dictionary containing the model contents from a .hdf5.gz file.
+
+        Parameters
+        ----------
+        fname : str
+            Full path to the .hdf5.gz file to load the :class:`Tidy3dBaseModel` from.
+        group_path : str, optional
+            Path to a group inside the file to selectively load a sub-element of the model only.
+        custom_decoders : List[Callable]
+            List of functions accepting
+            (fname: str, group_path: str, model_dict: dict, key: str, value: Any) that store the
+            value in the model dict after a custom decoding.
+
+        Returns
+        -------
+        dict
+            Dictionary containing the model.
+
+        Example
+        -------
+        >>> sim_dict = Simulation.dict_from_hdf5(fname='folder/sim.hdf5.gz') # doctest: +SKIP
+        """
+        file, extracted = tempfile.mkstemp(".hdf5")
+        os.close(file)
+        try:
+            extract_gzip_file(fname, extracted)
+            result = cls.dict_from_hdf5(
+                extracted, group_path=group_path, custom_decoders=custom_decoders
+            )
+        finally:
+            os.unlink(extracted)
+
+        return result
+
+    @classmethod
+    def from_hdf5_gz(
+        cls,
+        fname: str,
+        group_path: str = "",
+        custom_decoders: List[Callable] = None,
+        **parse_obj_kwargs,
+    ) -> Tidy3dBaseModel:
+        """Loads :class:`Tidy3dBaseModel` instance to .hdf5.gz file.
+
+        Parameters
+        ----------
+        fname : str
+            Full path to the .hdf5.gz file to load the :class:`Tidy3dBaseModel` from.
+        group_path : str, optional
+            Path to a group inside the file to selectively load a sub-element of the model only.
+            Starting `/` is optional.
+        custom_decoders : List[Callable]
+            List of functions accepting
+            (fname: str, group_path: str, model_dict: dict, key: str, value: Any) that store the
+            value in the model dict after a custom decoding.
+        **parse_obj_kwargs
+            Keyword arguments passed to pydantic's ``parse_obj`` method.
+
+        Example
+        -------
+        >>> simulation = Simulation.from_hdf5_gz(fname='folder/sim.hdf5.gz') # doctest: +SKIP
+        """
+
+        group_path = cls._construct_group_path(group_path)
+        model_dict = cls.dict_from_hdf5_gz(
+            fname=fname, group_path=group_path, custom_decoders=custom_decoders
+        )
+        return cls.parse_obj(model_dict, **parse_obj_kwargs)
+
+    def to_hdf5_gz(self, fname: str, custom_encoders: List[Callable] = None) -> None:
+        """Exports :class:`Tidy3dBaseModel` instance to .hdf5.gz file.
+
+        Parameters
+        ----------
+        fname : str
+            Full path to the .hdf5.gz file to save the :class:`Tidy3dBaseModel` to.
+        custom_encoders : List[Callable]
+            List of functions accepting (fname: str, group_path: str, value: Any) that take
+            the ``value`` supplied and write it to the hdf5 ``fname`` at ``group_path``.
+
+        Example
+        -------
+        >>> simulation.to_hdf5_gz(fname='folder/sim.hdf5.gz') # doctest: +SKIP
+        """
+
+        file, decompressed = tempfile.mkstemp(".hdf5")
+        os.close(file)
+        try:
+            self.to_hdf5(decompressed, custom_encoders=custom_encoders)
+            compress_file_to_gzip(decompressed, fname)
+        finally:
+            os.unlink(decompressed)
+
     def __lt__(self, other):
         """define < for getting unique indices based on hash."""
         return hash(self) < hash(other)
@@ -535,12 +762,69 @@ class Tidy3dBaseModel(pydantic.BaseModel):
 
     def __eq__(self, other):
         """Define == for two Tidy3DBaseModels."""
-        return self._json_string == other._json_string
+        if other is None:
+            return False
+
+        def check_equal(dict1: dict, dict2: dict) -> bool:
+            """Check if two dictionaries are equal, with special handlings."""
+
+            # if different keys, automatically fail
+            if not dict1.keys() == dict2.keys():
+                return False
+
+            # loop through elements in each dict
+            for key in dict1.keys():
+                val1 = dict1[key]
+                val2 = dict2[key]
+
+                # if one of val1 or val2 is None (exclusive OR)
+                if (val1 is None) != (val2 is None):
+                    return False
+
+                # convert tuple to dict to use this recursive function
+                if isinstance(val1, tuple) or isinstance(val2, tuple):
+                    val1 = dict(zip(range(len(val1)), val1))
+                    val2 = dict(zip(range(len(val2)), val2))
+
+                # if dictionaries, recurse
+                if isinstance(val1, dict) or isinstance(val2, dict):
+                    are_equal = check_equal(val1, val2)
+                    if not are_equal:
+                        return False
+
+                # if numpy arrays, use numpy to do equality check
+                elif isinstance(val1, np.ndarray) or isinstance(val2, np.ndarray):
+                    if not np.array_equal(val1, val2):
+                        return False
+
+                # everything else
+                else:
+                    # note: this logic is because != is handled differently in DataArrays apparently
+                    if not val1 == val2:
+                        return False
+
+            return True
+
+        return check_equal(self.dict(), other.dict())
 
     @cached_property
     def _json_string(self) -> str:
         """Returns string representation of a :class:`Tidy3dBaseModel`.
 
+        Returns
+        -------
+        str
+            Json-formatted string holding :class:`Tidy3dBaseModel` data.
+        """
+        return self._json()
+
+    def _json(self, indent=INDENT, exclude_unset=False, **kwargs) -> str:
+        """Overwrites the model ``json`` representation with some extra customized handling.
+
+        Parameters
+        -----------
+        **kwargs : kwargs passed to `self.json()`
+
         Returns
         -------
         str
@@ -548,19 +832,16 @@ class Tidy3dBaseModel(pydantic.BaseModel):
         """
 
         def make_json_compatible(json_string: str) -> str:
-            """Makes the string compatiable with json standards, notably for infinity."""
+            """Makes the string compatible with json standards, notably for infinity."""
 
             tmp_string = "<<TEMPORARY_INFINITY_STRING>>"
             json_string = json_string.replace("-Infinity", tmp_string)
             json_string = json_string.replace("Infinity", '"Infinity"')
             return json_string.replace(tmp_string, '"-Infinity"')
 
-        json_string = self.json(indent=INDENT, exclude_unset=False)
+        json_string = self.json(indent=indent, exclude_unset=exclude_unset, **kwargs)
         json_string = make_json_compatible(json_string)
         return json_string
-        # json_dict = json.loads(json_string)
-
-        # return json.dumps(json_dict)
 
     @classmethod
     def add_type_field(cls) -> None:
@@ -596,13 +877,12 @@ class Tidy3dBaseModel(pydantic.BaseModel):
         # create the list of parameters (arguments) for the model
         doc += "\n\n    Parameters\n    ----------\n"
         for field_name, field in cls.__fields__.items():
-
             # ignore the type tag
             if field_name == TYPE_TAG_STR:
                 continue
 
             # get data type
-            data_type = field._type_display()  # pylint:disable=protected-access
+            data_type = field._type_display()
 
             # get default values
             default_val = field.get_default()
@@ -645,3 +925,34 @@ class Tidy3dBaseModel(pydantic.BaseModel):
 
         doc += "\n"
         cls.__doc__ = doc
+
+    def get_submodels_by_hash(self) -> Dict[int, List[Union[str, Tuple[str, int]]]]:
+        """Return a dictionary of this object's sub-models indexed by their hash values."""
+        fields = {}
+        for key in self.__fields__:
+            field = getattr(self, key)
+
+            if isinstance(field, Tidy3dBaseModel):
+                hash_ = hash(field)
+                if hash_ not in fields:
+                    fields[hash_] = []
+                fields[hash_].append(key)
+
+            # Do we need to consider np.ndarray here?
+            elif isinstance(field, (list, tuple, np.ndarray)):
+                for index, sub_field in enumerate(field):
+                    if isinstance(sub_field, Tidy3dBaseModel):
+                        hash_ = hash(sub_field)
+                        if hash_ not in fields:
+                            fields[hash_] = []
+                        fields[hash_].append((key, index))
+
+            elif isinstance(field, dict):
+                for index, sub_field in field.items():
+                    if isinstance(sub_field, Tidy3dBaseModel):
+                        hash_ = hash(sub_field)
+                        if hash_ not in fields:
+                            fields[hash_] = []
+                        fields[hash_].append((key, index))
+
+        return fields
