@@ -4,7 +4,7 @@ from __future__ import annotations
 from abc import ABC
 from typing import Union, Tuple, List
 
-import pydantic as pd
+import pydantic.v1 as pd
 import numpy as np
 
 from .base import Tidy3dBaseModel, cached_property
@@ -27,6 +27,14 @@ class BoundaryEdge(ABC, Tidy3dBaseModel):
 class Periodic(BoundaryEdge):
     """Periodic boundary condition class."""
 
+    @property
+    def bloch_vec(self):
+        """Periodic boundaries are effectively Bloch boundaries with ``bloch_vec == 0``.
+        In practice, periodic boundaries do not force the use of complex fields, while Bloch
+        boundaries do, even with ``bloch_vec == 0``. Thus, it is more efficient to use periodic.
+        """
+        return 0
+
 
 # PEC keyword
 class PECBoundary(BoundaryEdge):
@@ -50,6 +58,13 @@ class BlochBoundary(BoundaryEdge):
     Example
     -------
     >>> bloch = BlochBoundary(bloch_vec=1)
+
+    See Also
+    --------
+
+    **Notebooks**:
+        * `Defining a total-field scattered-field (TFSF) plane wave source <../../notebooks/TFSF.html>`_
+        * `Multilevel blazed diffraction grating <../../notebooks/GratingEfficiency.html>`_
     """
 
     bloch_vec: float = pd.Field(
@@ -62,7 +77,7 @@ class BlochBoundary(BoundaryEdge):
 
     @cached_property
     def bloch_phase(self) -> Complex:
-        """Returns the forward phase factor associated with `bloch_vec`."""
+        """Returns the forward phase factor associated with ``bloch_vec``."""
         return np.exp(1j * 2.0 * np.pi * self.bloch_vec)
 
     @classmethod
@@ -97,19 +112,19 @@ class BlochBoundary(BoundaryEdge):
         >>> from tidy3d import GaussianPulse, PlaneWave, inf
         >>> pulse = GaussianPulse(freq0=200e12, fwidth=20e12)
         >>> pw_source = PlaneWave(
-        ...     size=(inf,inf,0), source_time=pulse, direction='+', angle_theta=1.5, angle_phi=0.3)
+        ...     size=(inf,inf,0), source_time=pulse, direction='+', angle_theta=0.2, angle_phi=0.3)
         >>> bloch = BlochBoundary.from_source(source=pw_source, domain_size=5, axis=0)
         """
 
         if not isinstance(source, BlochSourceType.__args__):
             raise SetupError(
-                "The `source` parameter must be `GaussianBeam`, `ModeSource`, `PlaneWave`, "
-                "or `TFSF` in order to define a Bloch boundary condition."
+                "The 'source' parameter must be 'GaussianBeam', 'ModeSource', 'PlaneWave', "
+                "or 'TFSF' in order to define a Bloch boundary condition."
             )
 
         if axis == source.injection_axis:
             raise SetupError(
-                "Bloch boundary axis must be orthogonal to the injection axis of `source`."
+                "Bloch boundary axis must be orthogonal to the injection axis of 'source'."
             )
 
         if medium is None:
@@ -260,9 +275,105 @@ class AbsorberSpec(BoundaryEdge):
 class PML(AbsorberSpec):
     """Specifies a standard PML along a single dimension.
 
+    Notes
+    ------
+
+        **1D Model Illustration**
+
+        Consider a transformed wave equation in the :math:`x` dimension below _`[1]`:
+
+        .. math::
+
+             \\left( \\left( \\frac{1}{s(x)} \\frac{\\delta}{\\delta x} \\right)^2 - \\frac{1}{c^2} \\frac{\\delta^2}{\\delta t^2} \\right) E = 0
+
+        where the wave stretch factor :math:`s(x)` depends on the PML boundary position in the :math:`x` dimension.
+
+        .. TODO what is x at 0?
+
+        .. math::
+
+             s(x) = \\left \\{
+                        \\begin{array}{lr}
+                            1, & \\text{for } x < 0 \\\\
+                            1 - \\frac{\\sigma}{i \\omega \\epsilon_0}, & \\text{for } x > 0
+                        \\end{array}
+                    \\right \\}
+
+        The wave equation can be solved and plotted accordingly as a function of the :math:`x` dimension.
+
+        .. math::
+
+             E(x) = \\left \\{
+                        \\begin{array}{lr}
+                            e^{i(kx - \\omega t)}, & \\text{for } x < 0 \\\\
+                            e^{i(kx - \\omega t)} \\times e^{-\\frac{\\sigma x}{c \\epsilon_0}} & \\text{for } x > 0
+                        \\end{array}
+                    \\right \\}
+
+        Hence, we see how this PML stretch factor induces frequency-independent exponential attentation and no
+        reflection after the boundary at :math:`x=0`.
+
+        .. image:: ../../_static/img/pml_boundary.png
+
+        .. TODO make this image better
+
+        **Usage Caveats**
+
+        A perfectly matched layer (PML) is the most commonly used boundary condition in FDTD simulations to truncate
+        a simulation domain and absorb outgoing radiation. However, many divergence issues are associated with the
+        use of PML. One of the most common causes of a diverged simulation is structures inserted into PML at an angle.
+
+        .. TODO links to absorber boundaries
+
+        .. image:: ../../notebooks/img/diverged-fdtd-simulation.png
+
+        Incorporating a dispersive material into the PML can also cause simulation divergence in certain scenarios.
+        If your simulation lacks any structures inserted into the PML at an angle, but includes dispersive material
+        in PML, it is advisable to substitute a nondispersive material for the dispersive material. Alternatively,
+        if dispersion is necessary, switching from the :class:`PML` to :class:`Absorber` can effectively address the
+        issue.
+
+        The PML can effectively absorb outgoing radiation with minimum reflection as if the radiation just propagates
+        into the free space. However, it’s important to keep in mind that the PML only absorbs propagating fields. For
+        evanescent fields, the PML can act as an amplification medium and cause a simulation to diverge. In Tidy3D,
+        a warning will appear if the distance between a structure is smaller than half of a wavelength to prevent
+        evanescent fields from leaking into PML. In most cases, the evanescent field will naturally die off within
+        half a wavelength, but in some instances, a larger distance may be required.
+
+        .. image:: ../../notebooks/img/diverged-fdtd-simulation1.png
+
+
+        **References**
+
+        .. [1]  W.C. Chew and W.H. Weedon, Microwave and Optical Tech. Lett., 7 (13), 599,1994; S. Johnson, arXiv 2108.05348, 2021
+        .. [2]  Antonios Giannopoulos, IEEE Transactions on Antennas and Propagation, 56(9), 2995, 2008
+
+    Note
+    ----
+
+        For best results, structures that intersect with the PML or simulation edges should extend extend all the way
+        through. In many such cases, an “infinite” size ``td.inf`` can be used to define the size along that dimension.
+
     Example
     -------
     >>> pml = PML(num_layers=10)
+
+    See Also
+    --------
+
+    :class:`StablePML`:
+         This PML deals handles possibly divergent simulations better, but at the expense of more layers.
+
+    :class:`Absorber`:
+         Specifies an adiabatic absorber along a single dimension.
+
+    **Notebooks:**
+        * `How to troubleshoot a diverged FDTD simulation <../../notebooks/DivergedFDTDSimulation.html>`_
+
+    **Lectures:**
+        * `Using FDTD to Compute a Transmission Spectrum <https://www.flexcompute.com/fdtd101/Lecture-2-Using-FDTD-to-Compute-a-Transmission-Spectrum/>`__
+        * `Introduction to perfectly matched layer (PML) tutorial <https://www.flexcompute.com/fdtd101/Lecture-6-Introduction-to-perfectly-matched-layer/>`__
+
     """
 
     num_layers: pd.NonNegativeInt = pd.Field(
@@ -280,11 +391,26 @@ class PML(AbsorberSpec):
 
 class StablePML(AbsorberSpec):
     """Specifies a 'stable' PML along a single dimension.
-    This PML deals handles possbly divergent simulations better, but at the expense of more layers.
+    This PML deals handles possibly divergent simulations better, but at the expense of more layers.
 
     Example
     -------
     >>> pml = StablePML(num_layers=40)
+
+    See Also
+    --------
+
+    :class:`PML`:
+         A standard PML along a single dimension.
+
+    :class:`Absorber`:
+         Specifies an adiabatic absorber along a single dimension.
+
+    **Notebooks:**
+        * `How to troubleshoot a diverged FDTD simulation <../../notebooks/DivergedFDTDSimulation.html>`_
+
+    **Lectures:**
+        * `Introduction to perfectly matched layer (PML) tutorial <https://www.flexcompute.com/fdtd101/Lecture-6-Introduction-to-perfectly-matched-layer/>`__
     """
 
     num_layers: pd.NonNegativeInt = pd.Field(
@@ -300,12 +426,41 @@ class StablePML(AbsorberSpec):
 
 class Absorber(AbsorberSpec):
     """Specifies an adiabatic absorber along a single dimension.
-    This absorber is well-suited for dispersive materials
-    intersecting with absorbing edges of the simulation at the expense of more layers.
+
+    Notes
+    -----
+
+        This absorber is well-suited for dispersive materials intersecting with absorbing edges of the simulation at the
+        expense of more layers.
+
+        **Usage Caveats**
+
+        Using absorber boundary is often a good remedy to resolve divergence issues related to :class:`PML`. The
+        adiabatic absorber is a multilayer system with gradually increasing conductivity. The absorber usually has a
+        larger undesired reflection compared to :class:`PML`. In practice, this small difference rarely matters,
+        but is important to understand for simulations that require high accuracy.
+
+        There are two possible sources for the reflection from absorbers. The first, and more common one, is that the
+        ramping up of the conductivity is not sufficiently slow, which can be remedied by increasing the number of
+        absorber layers (40 by default). The second one is that the absorption is not high enough, such that the
+        light reaches the :class:`PEC` boundary at the end of the :class:`Absorber`, travels back through it,
+        and is still not fully attenuated before re-entering the simulation region. If this is the case, increasing
+        the maximum conductivity :class:`AbsorberParams` can help. In both cases, changing the order of the scaling
+        of the conductivity (:attr:`tidy3d.AbsorberParams.sigma_order`) can also have an effect, but this is a more
+        advanced setting that we typically do not recommend modifying.
 
     Example
     -------
     >>> pml = Absorber(num_layers=40)
+
+    See Also
+    --------
+
+    :class:`PML`:
+         A standard PML along a single dimension.
+
+    **Notebooks:**
+        * `How to troubleshoot a diverged FDTD simulation <../../notebooks/DivergedFDTDSimulation.html>`_
     """
 
     num_layers: pd.NonNegativeInt = pd.Field(
@@ -335,11 +490,31 @@ BoundaryEdgeType = Union[
 
 
 class Boundary(Tidy3dBaseModel):
-    """Boundary conditions at the minus and plus extents along a dimension
+    """Boundary conditions at the minus and plus extents along a dimension.
+
+    Notes
+    -----
+
+         To specify individual boundary conditions along different dimensions, instead of :class:`BoundarySpec`,
+         this class is used, which defines the ``plus`` and ``minus`` boundaries along a single
+         dimension.
 
     Example
     -------
     >>> boundary = Boundary(plus = PML(), minus = PECBoundary())
+
+    See Also
+    --------
+
+    :class:`BoundarySpec`
+        Specifies boundary conditions on each side of the domain and along each dimension.
+
+    :class:`PML`
+        A standard PML along a single dimension.
+
+    **Notebooks**
+        * `Setting up boundary conditions <../../notebooks/BoundaryConditions.html>`_
+        * `Multilevel blazed diffraction grating <../../notebooks/GratingEfficiency.html>`_
     """
 
     plus: BoundaryEdgeType = pd.Field(
@@ -457,7 +632,7 @@ class Boundary(Tidy3dBaseModel):
         >>> from tidy3d import GaussianPulse, PlaneWave, inf
         >>> pulse = GaussianPulse(freq0=200e12, fwidth=20e12)
         >>> pw_source = PlaneWave(
-        ...     size=(inf,inf,0), source_time=pulse, direction='+', angle_theta=1.5, angle_phi=0.3)
+        ...     size=(inf,inf,0), source_time=pulse, direction='+', angle_theta=0.2, angle_phi=0.3)
         >>> bloch = Boundary.bloch_from_source(source=pw_source, domain_size=5, axis=0)
         """
         plus = BlochBoundary.from_source(
@@ -555,13 +730,45 @@ class Boundary(Tidy3dBaseModel):
 
 
 class BoundarySpec(Tidy3dBaseModel):
-    """Specifies boundary conditions on each side of the domain and along each dimension."""
+    """
+    Specifies boundary conditions on each side of the domain and along each dimension.
+
+    Notes
+    -----
+
+        This :class:`BoundarySpec` object defines the boundary conditions applied on each of the 6 domain edges,
+        and is provided as an input to the simulation.
+
+        A :class:`BoundarySpec` consists of three :class:`Boundary` objects, each defining the boundaries on the plus
+        and minus side of each dimension. In most cases, one just wants to specify whether there are absorbing
+        :class:`PML` layers along any of the ``x``, ``y``, ``z`` dimensions. By default, ``tidy3d`` simulations have
+        :class:`PML` boundaries on all sides.
+
+        If we want to explicitly set the boundaries, we can use the :attr:`tidy3d.BoundarySpec.all_sides` method.
+        This can be used to set any type of boundary condition on all sides of the simulation. We can also set
+        :class:`PML` on specified sides only by calling the :attr:`tidy3d.BoundarySpec.pml` method, e.g. ``BoundarySpec.pml(
+        x=False, y=False, z=False)``. This will put :class:`PML` along the dimensions defined as ``True``,
+        and set periodic boundaries along the other dimensions.
+
+
+    See Also
+    --------
+
+    :class:`Boundary`
+        Boundary conditions at the minus and plus extents along a dimension.
+
+    **Notebooks**
+        * `How to troubleshoot a diverged FDTD simulation <../../notebooks/DivergedFDTDSimulation.html>`_
+
+    **Lectures**
+        * `Using FDTD to Compute a Transmission Spectrum <https://www.flexcompute.com/fdtd101/Lecture-2-Using-FDTD-to-Compute-a-Transmission-Spectrum/>`__
+    """
 
     x: Boundary = pd.Field(
         Boundary(),
         title="Boundary condition along x.",
         description="Boundary condition on the plus and minus sides along the x axis. "
-        "If `None`, periodic boundaries are applied. Default will change to PML in 2.0 "
+        "If ``None``, periodic boundaries are applied. Default will change to PML in 2.0 "
         "so explicitly setting the boundaries is recommended.",
     )
 
@@ -569,7 +776,7 @@ class BoundarySpec(Tidy3dBaseModel):
         Boundary(),
         title="Boundary condition along y.",
         description="Boundary condition on the plus and minus sides along the y axis. "
-        "If `None`, periodic boundaries are applied. Default will change to PML in 2.0 "
+        "If ``None``, periodic boundaries are applied. Default will change to PML in 2.0 "
         "so explicitly setting the boundaries is recommended.",
     )
 
@@ -577,7 +784,7 @@ class BoundarySpec(Tidy3dBaseModel):
         Boundary(),
         title="Boundary condition along z.",
         description="Boundary condition on the plus and minus sides along the z axis. "
-        "If `None`, periodic boundaries are applied. Default will change to PML in 2.0 "
+        "If ``None``, periodic boundaries are applied. Default will change to PML in 2.0 "
         "so explicitly setting the boundaries is recommended.",
     )
 
