@@ -1,50 +1,99 @@
 """ Simulation Level Data """
 from __future__ import annotations
-from typing import Dict, Callable, Tuple
+from typing import Callable, Tuple, Union
 
+import pathlib
 import xarray as xr
-import pydantic as pd
+import pydantic.v1 as pd
 import numpy as np
+import h5py
+import json
 
 from .monitor_data import MonitorDataTypes, MonitorDataType, AbstractFieldData, FieldTimeData
-from ..base import Tidy3dBaseModel
 from ..simulation import Simulation
-from ..boundary import BlochBoundary
-from ..source import TFSF
 from ..types import Ax, Axis, annotate_type, FieldVal, PlotScale, ColormapType
 from ..viz import equal_aspect, add_ax_if_none
-from ...exceptions import DataError, Tidy3dKeyError, ValidationError
+from ...exceptions import DataError, Tidy3dKeyError
 from ...log import log
+from ..base import JSON_TAG
+
+from ..base_sim.data.sim_data import AbstractSimulationData
 
 
 DATA_TYPE_MAP = {data.__fields__["monitor"].type_: data for data in MonitorDataTypes}
 
+# maps monitor type (string) to the class of the corresponding data
+DATA_TYPE_NAME_MAP = {val.__fields__["monitor"].type_.__name__: val for val in MonitorDataTypes}
 
-class SimulationData(Tidy3dBaseModel):
+
+class SimulationData(AbstractSimulationData):
     """Stores data from a collection of :class:`.Monitor` objects in a :class:`.Simulation`.
 
-    Example
-    -------
-    >>> from tidy3d import ModeSpec, GridSpec, ScalarFieldDataArray, FieldMonitor, FieldData
+    Notes
+    -----
+
+        The ``SimulationData`` objects store a copy of the original :class:`.Simulation`:, so it can be recovered if the
+        ``SimulationData`` is loaded in a new session and the :class:`.Simulation` is no longer in memory.
+
+        More importantly, the ``SimulationData`` contains a reference to the data for each of the monitors within the
+        original :class:`.Simulation`. This data can be accessed directly using the name given to the monitors initially.
+
+    Examples
+    --------
+
+    Standalone example:
+
+    >>> import tidy3d as td
     >>> num_modes = 5
-    >>> x = [-1,1]
-    >>> y = [-2,0,2]
-    >>> z = [-3,-1,1,3]
+    >>> x = [-1,1,3]
+    >>> y = [-2,0,2,4]
+    >>> z = [-3,-1,1,3,5]
     >>> f = [2e14, 3e14]
-    >>> t = [0, 1e-12, 2e-12]
-    >>> mode_index = np.arange(num_modes)
-    >>> direction = ["+", "-"]
-    >>> coords = dict(x=x, y=y, z=z, f=f)
-    >>> scalar_field = ScalarFieldDataArray((1+1j) * np.random.random((2,3,4,2)), coords=coords)
-    >>> field_monitor = FieldMonitor(size=(2,4,6), freqs=[2e14, 3e14], name='field', fields=['Ex'])
+    >>> coords = dict(x=x[:-1], y=y[:-1], z=z[:-1], f=f)
+    >>> grid = td.Grid(boundaries=td.Coords(x=x, y=y, z=z))
+    >>> scalar_field = td.ScalarFieldDataArray((1+1j) * np.random.random((2,3,4,2)), coords=coords)
+    >>> field_monitor = td.FieldMonitor(
+    ...     size=(2,4,6),
+    ...     freqs=[2e14, 3e14],
+    ...     name='field',
+    ...     fields=['Ex'],
+    ...     colocate=True,
+    ... )
     >>> sim = Simulation(
     ...     size=(2, 4, 6),
-    ...     grid_spec=GridSpec(wavelength=1.0),
+    ...     grid_spec=td.GridSpec(wavelength=1.0),
     ...     monitors=[field_monitor],
     ...     run_time=2e-12,
+    ...     sources=[
+    ...         td.UniformCurrentSource(
+    ...             size=(0, 0, 0),
+    ...             center=(0, 0.5, 0),
+    ...             polarization="Hx",
+    ...             source_time=td.GaussianPulse(
+    ...                 freq0=2e14,
+    ...                 fwidth=4e13,
+    ...             ),
+    ...         )
+    ...     ],
     ... )
-    >>> field_data = FieldData(monitor=field_monitor, Ex=scalar_field)
-    >>> sim_data = SimulationData(simulation=sim, data=(field_data,))
+    >>> field_data = td.FieldData(monitor=field_monitor, Ex=scalar_field, grid_expanded=grid)
+    >>> sim_data = td.SimulationData(simulation=sim, data=(field_data,))
+
+    To save and load the :class:`SimulationData` object.
+
+    .. code-block:: python
+
+        sim_data.to_file(fname='path/to/file.hdf5') # Save a SimulationData object to a HDF5 file
+        sim_data = SimulationData.from_file(fname='path/to/file.hdf5') # Load a SimulationData object from a HDF5 file.
+
+    See Also
+    --------
+
+    **Notebooks:**
+        * `Quickstart <../../notebooks/StartHere.html>`_: Usage in a basic simulation flow.
+        * `Performing visualization of simulation data <../../notebooks/VizData.html>`_
+        * `Advanced monitor data manipulation and visualization <../../notebooks/XarrayTutorial.html>`_
+
     """
 
     simulation: Simulation = pd.Field(
@@ -60,45 +109,12 @@ class SimulationData(Tidy3dBaseModel):
         "associated with the monitors of the original :class:`.Simulation`.",
     )
 
-    log: str = pd.Field(
-        None,
-        title="Solver Log",
-        description="A string containing the log information from the simulation run.",
-    )
-
     diverged: bool = pd.Field(
         False,
         title="Diverged",
         description="A boolean flag denoting whether the simulation run diverged.",
     )
 
-    def __getitem__(self, monitor_name: str) -> MonitorDataType:
-        """Get a :class:`.MonitorData` by name. Apply symmetry if applicable."""
-        monitor_data = self.monitor_data[monitor_name]
-        return monitor_data.symmetry_expanded_copy
-
-    @property
-    def monitor_data(self) -> Dict[str, MonitorDataType]:
-        """Dictionary mapping monitor name to its associated :class:`.MonitorData`."""
-        return {monitor_data.monitor.name: monitor_data for monitor_data in self.data}
-
-    @pd.validator("data", always=True)
-    def data_monitors_match_sim(cls, val, values):
-        """Ensure each MonitorData in ``.data`` corresponds to a monitor in ``.simulation``."""
-        sim = values.get("simulation")
-        if sim is None:
-            raise ValidationError("Simulation.simulation failed validation, can't validate data.")
-        for mnt_data in val:
-            try:
-                monitor_name = mnt_data.monitor.name
-                sim.get_monitor_by_name(monitor_name)
-            except Tidy3dKeyError as exc:
-                raise DataError(
-                    f"Data with monitor name {monitor_name} supplied "
-                    "but not found in the Simulation"
-                ) from exc
-        return val
-
     @property
     def final_decay_value(self) -> float:
         """Returns value of the field decay at the final time step."""
@@ -108,7 +124,7 @@ class SimulationData(Tidy3dBaseModel):
                 "No log string in the SimulationData object, can't find final decay value."
             )
         lines = log_str.split("\n")
-        decay_lines = [l for l in lines if "field decay" in l]
+        decay_lines = [line for line in lines if "field decay" in line]
         final_decay = 1.0
         if len(decay_lines) > 0:
             final_decay_line = decay_lines[-1]
@@ -125,21 +141,17 @@ class SimulationData(Tidy3dBaseModel):
         source_time = source.source_time
         times = self.simulation.tmesh
         dt = self.simulation.dt
-        user_defined_phase = np.exp(1j * source_time.phase)
-
-        # get boundary information to determine whether to use complex fields
-        boundaries = self.simulation.boundary_spec.to_list
-        boundaries_1d = [boundary_1d for dim_boundary in boundaries for boundary_1d in dim_boundary]
-        complex_fields = any(isinstance(boundary, BlochBoundary) for boundary in boundaries_1d)
-        complex_fields = complex_fields and not isinstance(source, TFSF)
 
         # plug in mornitor_data frequency domain information
         def source_spectrum_fn(freqs):
             """Source amplitude as function of frequency."""
-            spectrum = source_time.spectrum(times, freqs, dt, complex_fields)
+            spectrum = source_time.spectrum(times, freqs, dt)
 
-            # remove user defined phase from normalization so its effect is present in the result
-            return spectrum * np.conj(user_defined_phase)
+            # Remove user defined amplitude and phase from the normalization
+            # such that they would still have an effect on the output fields.
+            # In other words, we are only normalizing out the arbitrary part of the spectrum
+            # that depends on things like freq0, fwidth and offset.
+            return spectrum / source_time.amplitude / np.exp(1j * source_time.phase)
 
         return source_spectrum_fn
 
@@ -178,13 +190,12 @@ class SimulationData(Tidy3dBaseModel):
         if not isinstance(mon_data, AbstractFieldData):
             raise DataError(
                 f"data for monitor '{monitor_name}' does not contain field data "
-                f"as it is a `{type(mon_data)}`."
+                f"as it is a '{type(mon_data)}'."
             )
         return mon_data
 
     def at_centers(self, field_monitor_name: str) -> xr.Dataset:
-        """return xarray.Dataset representation of field monitor data
-        co-located at Yee cell centers.
+        """Return xarray.Dataset representation of field monitor data colocated at Yee cell centers.
 
         Parameters
         ----------
@@ -194,26 +205,56 @@ class SimulationData(Tidy3dBaseModel):
         Returns
         -------
         xarray.Dataset
-            Dataset containing all of the fields in the data
-            interpolated to center locations on Yee grid.
+            Dataset containing all of the fields in the data interpolated to center locations on
+            the Yee grid.
         """
 
-        # get the data
         monitor_data = self.load_field_monitor(field_monitor_name)
+        return monitor_data.at_coords(monitor_data.colocation_centers)
 
-        # discretize the monitor and get center locations
-        sub_grid = self.simulation.discretize(monitor_data.monitor, extend=False)
-        centers = sub_grid.centers
+    def _at_boundaries(self, monitor_data: xr.Dataset) -> xr.Dataset:
+        """Return xarray.Dataset representation of field monitor data colocated at Yee cell
+        boundaries.
 
-        # pass coords if each of the scalar field data have more than one coordinate along a dim
-        xyz_kwargs = {}
-        for dim, centers in zip("xyz", (centers.x, centers.y, centers.z)):
-            scalar_data = list(monitor_data.field_components.values())
-            coord_lens = [len(data.coords[dim]) for data in scalar_data]
-            if all(ncoords > 1 for ncoords in coord_lens):
-                xyz_kwargs[dim] = centers
+        Parameters
+        ----------
+        monitor_data : xr.Dataset
+            Monitor data to be co-located.
 
-        return monitor_data.colocate(**xyz_kwargs)
+        Returns
+        -------
+        xarray.Dataset
+            Dataset containing all of the fields in the data interpolated to boundary locations on
+            the Yee grid.
+        """
+
+        if monitor_data.monitor.colocate:
+            # TODO: this still errors if monitor_data.colocate is allowed to be ``True`` in the
+            # adjoint plugin, and the monitor data is tracked in a gradient computation. It seems
+            # interpolating does something to the arrays that makes the JAX chain work.
+            return monitor_data.package_colocate_results(monitor_data.field_components)
+
+        # colocate to monitor grid boundaries
+        return monitor_data.at_coords(monitor_data.colocation_boundaries)
+
+    def at_boundaries(self, field_monitor_name: str) -> xr.Dataset:
+        """Return xarray.Dataset representation of field monitor data colocated at Yee cell
+        boundaries.
+
+        Parameters
+        ----------
+        field_monitor_name : str
+            Name of field monitor used in the original :class:`Simulation`.
+
+        Returns
+        -------
+        xarray.Dataset
+            Dataset containing all of the fields in the data interpolated to boundary locations on
+            the Yee grid.
+        """
+
+        # colocate to monitor grid boundaries
+        return self._at_boundaries(self.load_field_monitor(field_monitor_name))
 
     def get_poynting_vector(self, field_monitor_name: str) -> xr.Dataset:
         """return ``xarray.Dataset`` of the Poynting vector at Yee cell centers.
@@ -235,8 +276,8 @@ class SimulationData(Tidy3dBaseModel):
             DataArray containing the Poynting vector calculated based on the field components
             colocated at the center locations of the Yee grid.
         """
-
-        field_dataset = self.at_centers(field_monitor_name)
+        mon_data = self.load_field_monitor(field_monitor_name)
+        field_dataset = self._at_boundaries(mon_data)
 
         time_domain = isinstance(self.monitor_data[field_monitor_name], FieldTimeData)
 
@@ -261,47 +302,17 @@ class SimulationData(Tidy3dBaseModel):
                 else 0.5 * (e_1 * h_2.conj() - e_2 * h_1.conj())
             )
 
+            # 2D monitors have grid correction factors that can be different from 1. For Poynting,
+            # it is always the product of a primal-located field and dual-located field, so the
+            # total grid correction factor is the product of the two
+            grid_correction = mon_data.grid_dual_correction * mon_data.grid_primal_correction
+            poynting_components["S" + dim] *= grid_correction
+
         return xr.Dataset(poynting_components)
 
-    @staticmethod
-    def _field_component_value(field_component: xr.DataArray, val: FieldVal) -> xr.DataArray:
-        """return the desired value of a field component.
-
-        Parameter
-        ----------
-        field_component : xarray.DataArray
-            Field component from which to calculate the value.
-        val : Literal['real', 'imag', 'abs', 'abs^2', 'phase']
-            Which part of the field to return.
-
-        Returns
-        -------
-        xarray.DataArray
-            Value extracted from the field component.
-        """
-        if val == "real":
-            field_value = field_component.real
-            field_value.name = f"Re{{{field_component.name}}}"
-
-        elif val == "imag":
-            field_value = field_component.imag
-            field_value.name = f"Im{{{field_component.name}}}"
-
-        elif val == "abs":
-            field_value = np.abs(field_component)
-            field_value.name = f"|{field_component.name}|"
-
-        elif val == "abs^2":
-            field_value = np.abs(field_component) ** 2
-            field_value.name = f"|{field_component.name}|²"
-
-        elif val == "phase":
-            field_value = np.arctan2(field_component.imag, field_component.real)
-            field_value.name = f"∠{field_component.name}"
-
-        return field_value
-
-    def _get_scalar_field(self, field_monitor_name: str, field_name: str, val: FieldVal):
+    def _get_scalar_field(
+        self, field_monitor_name: str, field_name: str, val: FieldVal, phase: float = 0.0
+    ):
         """return ``xarray.DataArray`` of the scalar field of a given monitor at Yee cell centers.
 
         Parameters
@@ -312,6 +323,8 @@ class SimulationData(Tidy3dBaseModel):
             Name of the derived field component: one of `('E', 'H', 'S', 'Sx', 'Sy', 'Sz')`.
         val : Literal['real', 'imag', 'abs', 'abs^2', 'phase'] = 'real'
             Which part of the field to plot.
+        phase : float = 0.0
+            Optional phase to apply to result
 
         Returns
         -------
@@ -329,7 +342,9 @@ class SimulationData(Tidy3dBaseModel):
                     return self._field_component_value(derived_data, val)
                 raise Tidy3dKeyError(f"Poynting component {field_name} not available")
         else:
-            dataset = self.at_centers(field_monitor_name)
+            dataset = self.at_boundaries(field_monitor_name)
+
+        dataset = self.apply_phase(data=dataset, phase=phase)
 
         if field_name in ("E", "H", "S"):
             # Gather vector components
@@ -342,25 +357,33 @@ class SimulationData(Tidy3dBaseModel):
             field_components = (dataset[c] for c in required_components)
 
             # Apply the requested transformation
-            if val == "real":
+            val = val.lower()
+            if val in ("real", "re"):
                 derived_data = sum(f.real**2 for f in field_components) ** 0.5
                 derived_data.name = f"|Re{{{field_name}}}|"
 
-            elif val == "imag":
+            elif val in ("imag", "im"):
                 derived_data = sum(f.imag**2 for f in field_components) ** 0.5
                 derived_data.name = f"|Im{{{field_name}}}|"
 
             elif val == "abs":
-                derived_data = sum(np.abs(f) ** 2 for f in field_components) ** 0.5
+                derived_data = sum(abs(f) ** 2 for f in field_components) ** 0.5
                 derived_data.name = f"|{field_name}|"
 
             elif val == "abs^2":
-                derived_data = sum(np.abs(f) ** 2 for f in field_components)
-                derived_data.name = f"|{field_name}|²"
+                derived_data = sum(abs(f) ** 2 for f in field_components)
+                if hasattr(derived_data, "name"):
+                    derived_data.name = f"|{field_name}|²"
 
             elif val == "phase":
                 raise Tidy3dKeyError(f"Phase is not defined for complex vector {field_name}")
 
+            else:
+                raise Tidy3dKeyError(
+                    f"'val' of {val} not supported. "
+                    "Must be one of 'real', 'imag', 'abs', 'abs^2', or 'phase'."
+                )
+
             return derived_data
 
         raise Tidy3dKeyError(
@@ -386,7 +409,76 @@ class SimulationData(Tidy3dBaseModel):
             field_monitor_name=field_monitor_name, field_name="E", val="abs^2"
         )
 
-    # pylint: disable=too-many-arguments,too-many-locals,too-many-branches,too-many-statements
+    @classmethod
+    def mnt_data_from_file(cls, fname: str, mnt_name: str, **parse_obj_kwargs) -> MonitorDataType:
+        """Loads data for a specific monitor from a .hdf5 file with data for a ``SimulationData``.
+
+        Parameters
+        ----------
+        fname : str
+            Full path to an hdf5 file containing :class:`.SimulationData` data.
+        mnt_name : str, optional
+            ``.name`` of the monitor to load the data from.
+        **parse_obj_kwargs
+            Keyword arguments passed to either pydantic's ``parse_obj`` function when loading model.
+
+        Returns
+        -------
+        :class:`MonitorData`
+            Monitor data corresponding to the ``mnt_name`` type.
+
+        Example
+        -------
+        >>> field_data = SimulationData.from_file(fname='folder/data.hdf5', mnt_name="field") # doctest: +SKIP
+        """
+
+        if pathlib.Path(fname).suffix != ".hdf5":
+            raise ValueError("'mnt_data_from_file' only works with '.hdf5' files.")
+
+        # open file and ensure it has data
+        with h5py.File(fname) as f_handle:
+            if "data" not in f_handle:
+                raise ValueError(f"could not find data in the supplied file {fname}")
+
+            # get the monitor list from the json string
+            json_string = f_handle[JSON_TAG][()]
+            json_dict = json.loads(json_string)
+            monitor_list = json_dict["simulation"]["monitors"]
+
+            # loop through data
+            for monitor_index_str, _mnt_data in f_handle["data"].items():
+                # grab the monitor data for this data element
+                monitor_dict = monitor_list[int(monitor_index_str)]
+
+                # if a match on the monitor name
+                if monitor_dict["name"] == mnt_name:
+                    # try to grab the monitor data type
+                    monitor_type_str = monitor_dict["type"]
+                    if monitor_type_str not in DATA_TYPE_NAME_MAP:
+                        raise ValueError(f"Could not find data type '{monitor_type_str}'.")
+                    monitor_data_type = DATA_TYPE_NAME_MAP[monitor_type_str]
+
+                    # load the monitor data from the file using the group_path
+                    group_path = f"data/{monitor_index_str}"
+                    return monitor_data_type.from_file(
+                        fname, group_path=group_path, **parse_obj_kwargs
+                    )
+
+        raise ValueError(f"No monitor with name '{mnt_name}' found in data file.")
+
+    @staticmethod
+    def apply_phase(data: Union[xr.DataArray, xr.Dataset], phase: float = 0.0) -> xr.DataArray:
+        """Apply a phase to xarray data."""
+        if phase != 0.0:
+            if np.any(np.iscomplex(data.values)):
+                data *= np.exp(1j * phase)
+            else:
+                log.warning(
+                    f"Non-zero phase of {phase} specified but the data being plotted is "
+                    "real-valued. The phase will be ignored in the plot."
+                )
+        return data
+
     def plot_field(
         self,
         field_monitor_name: str,
@@ -394,13 +486,14 @@ class SimulationData(Tidy3dBaseModel):
         val: FieldVal = "real",
         scale: PlotScale = "lin",
         eps_alpha: float = 0.2,
+        phase: float = 0.0,
         robust: bool = True,
         vmin: float = None,
         vmax: float = None,
         ax: Ax = None,
         **sel_kwargs,
     ) -> Ax:
-        """Plot the field data for a monitor with simulation plot overlayed.
+        """Plot the field data for a monitor with simulation plot overlaid.
 
         Parameters
         ----------
@@ -408,9 +501,9 @@ class SimulationData(Tidy3dBaseModel):
             Name of :class:`.FieldMonitor`, :class:`.FieldTimeData`, or :class:`.ModeSolverData`
             to plot.
         field_name : str
-            Name of `field` component to plot (eg. `'Ex'`).
-            Also accepts `'E'` and `'H'` to plot the vector magnitudes of the electric and
-            magnetic fields, and `'S'` for the Poynting vector.
+            Name of ``field`` component to plot (eg. `'Ex'`).
+            Also accepts ``'E'`` and ``'H'`` to plot the vector magnitudes of the electric and
+            magnetic fields, and ``'S'`` for the Poynting vector.
         val : Literal['real', 'imag', 'abs', 'abs^2', 'phase'] = 'real'
             Which part of the field to plot.
         scale : Literal['lin', 'dB']
@@ -418,24 +511,27 @@ class SimulationData(Tidy3dBaseModel):
         eps_alpha : float = 0.2
             Opacity of the structure permittivity.
             Must be between 0 and 1 (inclusive).
+        phase : float = 0.0
+            Optional phase (radians) to apply to the fields.
+            Only has an effect on frequency-domain fields.
         robust : bool = True
             If True and vmin or vmax are absent, uses the 2nd and 98th percentiles of the data
             to compute the color limits. This helps in visualizing the field patterns especially
             in the presence of a source.
         vmin : float = None
-            The lower bound of data range that the colormap covers. If `None`, they are
+            The lower bound of data range that the colormap covers. If ``None``, they are
             inferred from the data and other keyword arguments.
         vmax : float = None
-            The upper bound of data range that the colormap covers. If `None`, they are
+            The upper bound of data range that the colormap covers. If ``None``, they are
             inferred from the data and other keyword arguments.
         ax : matplotlib.axes._subplots.Axes = None
             matplotlib axes to plot on, if not specified, one is created.
-        sel_kwargs : keyword arguments used to perform `.sel()` selection in the monitor data.
-            These kwargs can select over the spatial dimensions (`x`, `y`, `z`),
-            frequency or time dimensions (`f`, `t`) or `mode_index`, if applicable.
+        sel_kwargs : keyword arguments used to perform ``.sel()`` selection in the monitor data.
+            These kwargs can select over the spatial dimensions (``x``, ``y``, ``z``),
+            frequency or time dimensions (``f``, ``t``) or ``mode_index``, if applicable.
             For the plotting to work appropriately, the resulting data after selection must contain
             only two coordinates with len > 1.
-            Furthermore, these should be spatial coordinates (`x`, `y`, or `z`).
+            Furthermore, these should be spatial coordinates (``x``, ``y``, or ``z``).
 
         Returns
         -------
@@ -444,11 +540,10 @@ class SimulationData(Tidy3dBaseModel):
         """
 
         # get the DataArray corresponding to the monitor_name and field_name
-
         # deprecated intensity
         if field_name == "int":
             log.warning(
-                "'int' field name is deprecated and will be removed in the future. Plese use "
+                "'int' field name is deprecated and will be removed in the future. Please use "
                 "field_name='E' and val='abs^2' for the same effect."
             )
             field_name = "E"
@@ -456,7 +551,7 @@ class SimulationData(Tidy3dBaseModel):
 
         if field_name in ("E", "H") or field_name[0] == "S":
             # Derived fields
-            field_data = self._get_scalar_field(field_monitor_name, field_name, val)
+            field_data = self._get_scalar_field(field_monitor_name, field_name, val, phase=phase)
         else:
             # Direct field component (e.g. Ex)
             field_monitor_data = self.load_field_monitor(field_monitor_name)
@@ -464,11 +559,12 @@ class SimulationData(Tidy3dBaseModel):
                 raise DataError(f"field_name '{field_name}' not found in data.")
             field_component = field_monitor_data.field_components[field_name]
             field_component.name = field_name
+            field_component = self.apply_phase(data=field_component, phase=phase)
             field_data = self._field_component_value(field_component, val)
 
         if scale == "dB":
             if val == "phase":
-                log.warning("Ploting phase component in log scale masks the phase sign.")
+                log.warning("Plotting phase component in log scale masks the phase sign.")
             db_factor = {
                 ("S", "real"): 10,
                 ("S", "imag"): 10,
@@ -579,7 +675,6 @@ class SimulationData(Tidy3dBaseModel):
             ax=ax,
         )
 
-    # pylint: disable=too-many-arguments,too-many-locals
     @equal_aspect
     @add_ax_if_none
     def plot_scalar_array(
@@ -595,7 +690,7 @@ class SimulationData(Tidy3dBaseModel):
         cmap_type: ColormapType = "divergent",
         ax: Ax = None,
     ) -> Ax:
-        """Plot the field data for a monitor with simulation plot overlayed.
+        """Plot the field data for a monitor with simulation plot overlaid.
 
         Parameters
         ----------
