@@ -4,7 +4,7 @@ from __future__ import annotations
 from typing import Dict, Tuple, Union, List
 import numpy as np
 import xarray as xr
-import pydantic
+import pydantic.v1 as pydantic
 
 from rich.progress import track
 
@@ -19,20 +19,35 @@ from .monitor import FieldMonitor, AbstractFieldProjectionMonitor, FieldProjecti
 from .monitor import FieldProjectionCartesianMonitor, FieldProjectionKSpaceMonitor
 from .types import Direction, Coordinate, ArrayComplex4D
 from .medium import MediumType
-from .base import Tidy3dBaseModel, cached_property
+from .base import Tidy3dBaseModel, cached_property, skip_if_fields_missing
 from ..exceptions import SetupError
 from ..constants import C_0, MICROMETER, ETA_0, EPSILON_0, MU_0
+from ..log import get_logging_console
 
 # Default number of points per wavelength in the background medium to use for resampling fields.
 PTS_PER_WVL = 10
 
 # Numpy float array and related array types
-# pylint: disable=invalid-name
+
 ArrayLikeN2F = Union[float, Tuple[float, ...], ArrayComplex4D]
 
 
 class FieldProjector(Tidy3dBaseModel):
-    """Projection of near fields to points on a given observation grid."""
+    """
+    Projection of near fields to points on a given observation grid.
+
+    .. TODO make images to illustrate this
+
+    See Also
+    --------
+
+    :class:`FieldProjectionAngleMonitor
+        :class:`Monitor` that samples electromagnetic near fields in the frequency domain
+        and projects them at given observation angles.`
+
+    **Notebooks**:
+        * `Performing near field to far field projections <../../notebooks/FieldProjections.html>`_
+    """
 
     sim_data: SimulationData = pydantic.Field(
         ...,
@@ -71,6 +86,7 @@ class FieldProjector(Tidy3dBaseModel):
     )
 
     @pydantic.validator("origin", always=True)
+    @skip_if_fields_missing(["surfaces"])
     def set_origin(cls, val, values):
         """Sets .origin as the average of centers of all surface monitors if not provided."""
         if val is None:
@@ -92,7 +108,7 @@ class FieldProjector(Tidy3dBaseModel):
         return self.surfaces[0].monitor.freqs
 
     @classmethod
-    def from_near_field_monitors(  # pylint:disable=too-many-arguments
+    def from_near_field_monitors(
         cls,
         sim_data: SimulationData,
         near_monitors: List[FieldMonitor],
@@ -139,7 +155,6 @@ class FieldProjector(Tidy3dBaseModel):
 
     @cached_property
     def currents(self):
-
         """Sets the surface currents."""
         sim_data = self.sim_data
         surfaces = self.surfaces
@@ -202,9 +217,7 @@ class FieldProjector(Tidy3dBaseModel):
         return currents
 
     @staticmethod
-    def _fields_to_currents(  # pylint:disable=too-many-locals
-        field_data: FieldData, surface: FieldProjectionSurface
-    ) -> FieldData:
+    def _fields_to_currents(field_data: FieldData, surface: FieldProjectionSurface) -> FieldData:
         """Returns surface current densities associated with a given :class:`.FieldData` object.
 
         Parameters
@@ -253,7 +266,6 @@ class FieldProjector(Tidy3dBaseModel):
         )
 
     @staticmethod
-    # pylint:disable=too-many-locals, too-many-arguments
     def _resample_surface_currents(
         currents: xr.Dataset,
         sim_data: SimulationData,
@@ -297,7 +309,6 @@ class FieldProjector(Tidy3dBaseModel):
         _, idx_uv = surface.monitor.pop_axis((0, 1, 2), axis=surface.axis)
 
         for idx in idx_uv:
-
             # pick sample points on the monitor and handle the possibility of an "infinite" monitor
             start = np.maximum(
                 surface.monitor.center[idx] - surface.monitor.size[idx] / 2.0,
@@ -309,7 +320,7 @@ class FieldProjector(Tidy3dBaseModel):
             )
 
             if pts_per_wavelength is None:
-                points = sim_data.simulation.grid.centers.to_list[idx]
+                points = sim_data.simulation.grid.boundaries.to_list[idx]
                 points[np.argwhere(points < start)] = start
                 points[np.argwhere(points > stop)] = stop
                 colocation_points[idx] = np.unique(points)
@@ -320,7 +331,7 @@ class FieldProjector(Tidy3dBaseModel):
                 colocation_points[idx] = points
 
         for idx, points in enumerate(colocation_points):
-            if (hasattr(points, "__len__") and len(points) == 1) or not hasattr(points, "__len__"):
+            if np.array(points).size == 1:
                 colocation_points[idx] = None
 
         currents = currents.colocate(*colocation_points)
@@ -336,7 +347,6 @@ class FieldProjector(Tidy3dBaseModel):
         """Trapezoidal integration in two dimensions."""
         return np.trapz(np.trapz(np.squeeze(function) * phase, pts_u, axis=0), pts_v, axis=0)
 
-    # pylint:disable=too-many-locals, too-many-arguments
     def _far_fields_for_surface(
         self,
         frequency: float,
@@ -344,6 +354,7 @@ class FieldProjector(Tidy3dBaseModel):
         phi: ArrayLikeN2F,
         surface: FieldProjectionSurface,
         currents: xr.Dataset,
+        medium: MediumType,
     ):
         """Compute far fields at an angle in spherical coordinates
         for a given set of surface currents and observation angles.
@@ -361,13 +372,14 @@ class FieldProjector(Tidy3dBaseModel):
             :class:`FieldProjectionSurface` object to use as source of near field.
         currents : xarray.Dataset
             xarray Dataset containing surface currents associated with the surface monitor.
+        medium : :class:`.MediumType`
+            Background medium through which to project fields.
 
         Returns
         -------
         tuple(numpy.ndarray[float], ...)
             ``Er``, ``Etheta``, ``Ephi``, ``Hr``, ``Htheta``, ``Hphi`` for the given surface.
         """
-
         pts = [currents[name].values for name in ["x", "y", "z"]]
 
         try:
@@ -396,14 +408,13 @@ class FieldProjector(Tidy3dBaseModel):
 
         phase = [None] * 3
         propagation_factor = -1j * AbstractFieldProjectionData.wavenumber(
-            medium=self.medium, frequency=frequency
+            medium=medium, frequency=frequency
         )
 
         def integrate_for_one_theta(i_th: int):
             """Perform integration for a given theta angle index"""
 
             for j_ph in np.arange(len(phi)):
-
                 phase[0] = np.exp(propagation_factor * pts[0] * sin_theta[i_th] * cos_phi[j_ph])
                 phase[1] = np.exp(propagation_factor * pts[1] * sin_theta[i_th] * sin_phi[j_ph])
                 phase[2] = np.exp(propagation_factor * pts[2] * cos_theta[i_th])
@@ -432,6 +443,7 @@ class FieldProjector(Tidy3dBaseModel):
             for i_th in track(
                 np.arange(len(theta)),
                 description=f"Processing surface monitor '{surface.monitor.name}'...",
+                console=get_logging_console(),
             ):
                 integrate_for_one_theta(i_th)
 
@@ -450,7 +462,7 @@ class FieldProjector(Tidy3dBaseModel):
         # Lphi  (8.34b)
         Lphi = -M[0] * sin_phi[None, :] + M[1] * cos_phi[None, :]
 
-        eta = ETA_0 / np.sqrt(self.medium.eps_model(frequency))
+        eta = ETA_0 / np.sqrt(medium.eps_model(frequency))
 
         Etheta = -(Lphi + eta * Ntheta)
         Ephi = Ltheta - eta * Nphi
@@ -461,6 +473,45 @@ class FieldProjector(Tidy3dBaseModel):
 
         return Er, Etheta, Ephi, Hr, Htheta, Hphi
 
+    @staticmethod
+    def apply_window_to_currents(
+        proj_monitor: AbstractFieldProjectionMonitor, currents: xr.Dataset
+    ) -> xr.Dataset:
+        """Apply windowing function to the surface currents."""
+        if proj_monitor.size.count(0.0) == 0:
+            return currents
+        if proj_monitor.window_size == (0, 0):
+            return currents
+
+        pts = [currents[name].values for name in ["x", "y", "z"]]
+
+        custom_bounds = [
+            [pts[i][0] for i in range(3)],
+            [pts[i][-1] for i in range(3)],
+        ]
+
+        window_size, window_minus, window_plus = proj_monitor.window_parameters(
+            custom_bounds=custom_bounds
+        )
+
+        new_currents = currents.copy(deep=True)
+        for dim, (dim_name, points) in enumerate(zip("xyz", pts)):
+            window_fn = proj_monitor.window_function(
+                points=points,
+                window_size=window_size,
+                window_minus=window_minus,
+                window_plus=window_plus,
+                dim=dim,
+            )
+            window_data = xr.DataArray(
+                window_fn,
+                dims=[dim_name],
+                coords=[points],
+            )
+            new_currents *= window_data
+
+        return new_currents
+
     def project_fields(
         self, proj_monitor: AbstractFieldProjectionMonitor
     ) -> AbstractFieldProjectionData:
@@ -509,17 +560,25 @@ class FieldProjector(Tidy3dBaseModel):
             np.zeros((1, len(theta), len(phi), len(freqs)), dtype=complex) for _ in field_names
         ]
 
-        k = AbstractFieldProjectionData.wavenumber(medium=self.medium, frequency=freqs)
+        medium = monitor.medium if monitor.medium else self.medium
+        k = AbstractFieldProjectionData.wavenumber(medium=medium, frequency=freqs)
         phase = np.atleast_1d(
             AbstractFieldProjectionData.propagation_phase(dist=monitor.proj_distance, k=k)
         )
 
         for surface in self.surfaces:
+            # apply windowing to currents
+            currents = self.apply_window_to_currents(monitor, self.currents[surface.monitor.name])
 
             if monitor.far_field_approx:
                 for idx_f, frequency in enumerate(freqs):
                     _fields = self._far_fields_for_surface(
-                        frequency, theta, phi, surface, self.currents[surface.monitor.name]
+                        frequency=frequency,
+                        theta=theta,
+                        phi=phi,
+                        surface=surface,
+                        currents=currents,
+                        medium=medium,
                     )
                     for field, _field in zip(fields, _fields):
                         field[..., idx_f] += _field * phase[idx_f]
@@ -532,10 +591,11 @@ class FieldProjector(Tidy3dBaseModel):
                 for (_theta, _phi), (i, j) in track(
                     iter_coords,
                     description=f"Processing surface monitor '{surface.monitor.name}'...",
+                    console=get_logging_console(),
                 ):
                     _x, _y, _z = monitor.sph_2_car(monitor.proj_distance, _theta, _phi)
                     _fields = self._fields_for_surface_exact(
-                        _x, _y, _z, surface, self.currents[surface.monitor.name]
+                        x=_x, y=_y, z=_z, surface=surface, currents=currents, medium=medium
                     )
                     for field, _field in zip(fields, _fields):
                         field[0, i, j, :] += _field
@@ -546,7 +606,7 @@ class FieldProjector(Tidy3dBaseModel):
             for name, field in zip(field_names, fields)
         }
         return FieldProjectionAngleData(
-            monitor=monitor, projection_surfaces=self.surfaces, medium=self.medium, **fields
+            monitor=monitor, projection_surfaces=self.surfaces, medium=medium, **fields
         )
 
     def _project_fields_cartesian(
@@ -577,7 +637,8 @@ class FieldProjector(Tidy3dBaseModel):
             np.zeros((len(x), len(y), len(z), len(freqs)), dtype=complex) for _ in field_names
         ]
 
-        wavenumber = AbstractFieldProjectionData.wavenumber(medium=self.medium, frequency=freqs)
+        medium = monitor.medium if monitor.medium else self.medium
+        wavenumber = AbstractFieldProjectionData.wavenumber(medium=medium, frequency=freqs)
 
         # Zip together all combinations of observation points for better progress tracking
         iter_coords = [
@@ -587,24 +648,35 @@ class FieldProjector(Tidy3dBaseModel):
             for k, _z in enumerate(z)
         ]
 
-        for (_x, _y, _z), (i, j, k) in track(iter_coords, description="Computing projected fields"):
+        for (_x, _y, _z), (i, j, k) in track(
+            iter_coords, description="Computing projected fields", console=get_logging_console()
+        ):
             r, theta, phi = monitor.car_2_sph(_x, _y, _z)
             phase = np.atleast_1d(
                 AbstractFieldProjectionData.propagation_phase(dist=r, k=wavenumber)
             )
 
             for surface in self.surfaces:
+                # apply windowing to currents
+                currents = self.apply_window_to_currents(
+                    monitor, self.currents[surface.monitor.name]
+                )
 
                 if monitor.far_field_approx:
                     for idx_f, frequency in enumerate(freqs):
                         _fields = self._far_fields_for_surface(
-                            frequency, theta, phi, surface, self.currents[surface.monitor.name]
+                            frequency=frequency,
+                            theta=theta,
+                            phi=phi,
+                            surface=surface,
+                            currents=currents,
+                            medium=medium,
                         )
                         for field, _field in zip(fields, _fields):
                             field[i, j, k, idx_f] += _field * phase[idx_f]
                 else:
                     _fields = self._fields_for_surface_exact(
-                        _x, _y, _z, surface, self.currents[surface.monitor.name]
+                        x=_x, y=_y, z=_z, surface=surface, currents=currents, medium=medium
                     )
                     for field, _field in zip(fields, _fields):
                         field[i, j, k, :] += _field
@@ -615,7 +687,7 @@ class FieldProjector(Tidy3dBaseModel):
             for name, field in zip(field_names, fields)
         }
         return FieldProjectionCartesianData(
-            monitor=monitor, projection_surfaces=self.surfaces, medium=self.medium, **fields
+            monitor=monitor, projection_surfaces=self.surfaces, medium=medium, **fields
         )
 
     def _project_fields_kspace(
@@ -642,7 +714,8 @@ class FieldProjector(Tidy3dBaseModel):
         field_names = ("Er", "Etheta", "Ephi", "Hr", "Htheta", "Hphi")
         fields = [np.zeros((len(ux), len(uy), 1, len(freqs)), dtype=complex) for _ in field_names]
 
-        k = AbstractFieldProjectionData.wavenumber(medium=self.medium, frequency=freqs)
+        medium = monitor.medium if monitor.medium else self.medium
+        k = AbstractFieldProjectionData.wavenumber(medium=medium, frequency=freqs)
         phase = np.atleast_1d(
             AbstractFieldProjectionData.propagation_phase(dist=monitor.proj_distance, k=k)
         )
@@ -650,15 +723,26 @@ class FieldProjector(Tidy3dBaseModel):
         # Zip together all combinations of observation points for better progress tracking
         iter_coords = [([_ux, _uy], [i, j]) for i, _ux in enumerate(ux) for j, _uy in enumerate(uy)]
 
-        for (_ux, _uy), (i, j) in track(iter_coords, description="Computing projected fields"):
+        for (_ux, _uy), (i, j) in track(
+            iter_coords, description="Computing projected fields", console=get_logging_console()
+        ):
             theta, phi = monitor.kspace_2_sph(_ux, _uy, monitor.proj_axis)
 
             for surface in self.surfaces:
+                # apply windowing to currents
+                currents = self.apply_window_to_currents(
+                    monitor, self.currents[surface.monitor.name]
+                )
 
                 if monitor.far_field_approx:
                     for idx_f, frequency in enumerate(freqs):
                         _fields = self._far_fields_for_surface(
-                            frequency, theta, phi, surface, self.currents[surface.monitor.name]
+                            frequency=frequency,
+                            theta=theta,
+                            phi=phi,
+                            surface=surface,
+                            currents=currents,
+                            medium=medium,
                         )
                         for field, _field in zip(fields, _fields):
                             field[i, j, 0, idx_f] += _field * phase[idx_f]
@@ -666,7 +750,7 @@ class FieldProjector(Tidy3dBaseModel):
                 else:
                     _x, _y, _z = monitor.sph_2_car(monitor.proj_distance, theta, phi)
                     _fields = self._fields_for_surface_exact(
-                        _x, _y, _z, surface, self.currents[surface.monitor.name]
+                        x=_x, y=_y, z=_z, surface=surface, currents=currents, medium=medium
                     )
                     for field, _field in zip(fields, _fields):
                         field[i, j, 0, :] += _field
@@ -682,12 +766,11 @@ class FieldProjector(Tidy3dBaseModel):
             for name, field in zip(field_names, fields)
         }
         return FieldProjectionKSpaceData(
-            monitor=monitor, projection_surfaces=self.surfaces, medium=self.medium, **fields
+            monitor=monitor, projection_surfaces=self.surfaces, medium=medium, **fields
         )
 
     """Exact projections"""
 
-    # pylint:disable=too-many-locals, too-many-arguments, too-many-statements, invalid-name
     def _fields_for_surface_exact(
         self,
         x: float,
@@ -695,6 +778,7 @@ class FieldProjector(Tidy3dBaseModel):
         z: float,
         surface: FieldProjectionSurface,
         currents: xr.Dataset,
+        medium: MediumType,
     ):
         """Compute projected fields in spherical coordinates at a given projection point on a
         Cartesian grid for a given set of surface currents using the exact homogeneous medium
@@ -712,6 +796,8 @@ class FieldProjector(Tidy3dBaseModel):
             :class:`FieldProjectionSurface` object to use as source of near field.
         currents : xarray.Dataset
             xarray Dataset containing surface currents associated with the surface monitor.
+        medium : :class:`.MediumType`
+            Background medium through which to project fields.
 
         Returns
         -------
@@ -719,13 +805,12 @@ class FieldProjector(Tidy3dBaseModel):
             ``Er``, ``Etheta``, ``Ephi``, ``Hr``, ``Htheta``, ``Hphi`` projected fields for
             each frequency.
         """
-
         freqs = np.array(self.frequencies)
         i_omega = 1j * 2.0 * np.pi * freqs[None, None, None, :]
-        wavenumber = AbstractFieldProjectionData.wavenumber(frequency=freqs, medium=self.medium)
+        wavenumber = AbstractFieldProjectionData.wavenumber(frequency=freqs, medium=medium)
         wavenumber = wavenumber[None, None, None, :]  # add space dimensions
 
-        eps_complex = self.medium.eps_model(frequency=freqs)
+        eps_complex = medium.eps_model(frequency=freqs)
         epsilon = EPSILON_0 * eps_complex[None, None, None, :]
 
         # source points
@@ -733,7 +818,7 @@ class FieldProjector(Tidy3dBaseModel):
 
         # transform the coordinate system so that the origin is at the source point
         # then the observation points in the new system are:
-        x_new, y_new, z_new = [pt_obs - pt_src for pt_src, pt_obs in zip(pts, [x, y, z])]
+        x_new, y_new, z_new = (pt_obs - pt_src for pt_src, pt_obs in zip(pts, [x, y, z]))
 
         # tangential source components to use
         idx_w, idx_uv = surface.monitor.pop_axis((0, 1, 2), axis=surface.axis)
@@ -827,16 +912,16 @@ class FieldProjector(Tidy3dBaseModel):
         F, curl_F, grad_div_F = potential_terms(M, epsilon)
 
         # assemble the electric field components (Taflove 8.24, 8.27)
-        e_x_integrand, e_y_integrand, e_z_integrand = [
+        e_x_integrand, e_y_integrand, e_z_integrand = (
             i_omega * (a + grad_div_a / (wavenumber**2)) - curl_f / epsilon
             for a, grad_div_a, curl_f in zip(A, grad_div_A, curl_F)
-        ]
+        )
 
         # assemble the magnetic field components (Taflove 8.25, 8.28)
-        h_x_integrand, h_y_integrand, h_z_integrand = [
+        h_x_integrand, h_y_integrand, h_z_integrand = (
             i_omega * (f + grad_div_f / (wavenumber**2)) + curl_a / MU_0
             for f, grad_div_f, curl_a in zip(F, grad_div_F, curl_A)
-        ]
+        )
 
         # integrate over the surface
         e_x = self.integrate_2d(e_x_integrand, 1.0, pts[idx_u], pts[idx_v])
