@@ -5,15 +5,15 @@ from abc import ABC, abstractmethod
 from typing import Tuple, List, Union
 
 import numpy as np
-import pydantic as pd
+import pydantic.v1 as pd
 
 from .grid import Coords1D, Coords, Grid
 from .mesher import GradedMesher, MesherType
 from ..base import Tidy3dBaseModel
-from ..types import Axis, Symmetry, annotate_type
+from ..types import Axis, Symmetry, annotate_type, TYPE_TAG_STR
 from ..source import SourceType
 from ..structure import Structure, StructureType
-from ..geometry import Box
+from ..geometry.base import Box
 from ...log import log
 from ...exceptions import SetupError
 from ...constants import MICROMETER, C_0, fp_eps
@@ -23,7 +23,7 @@ class GridSpec1d(Tidy3dBaseModel, ABC):
 
     """Abstract base class, defines 1D grid generation specifications."""
 
-    def make_coords(  # pylint:disable = too-many-arguments
+    def make_coords(
         self,
         axis: Axis,
         structures: List[StructureType],
@@ -68,7 +68,7 @@ class GridSpec1d(Tidy3dBaseModel, ABC):
             is_periodic=is_periodic,
         )
 
-        # incooperate symmetries
+        # incorporate symmetries
         if symmetry[axis] != 0:
             # Offset to center if symmetry present
             center = structures[0].geometry.center[axis]
@@ -135,11 +135,21 @@ class GridSpec1d(Tidy3dBaseModel, ABC):
 
 class UniformGrid(GridSpec1d):
 
-    """Uniform 1D grid.
+    """Uniform 1D grid. The most standard way to define a simulation is to use a constant grid size in each of the three directions.
 
     Example
     -------
     >>> grid_1d = UniformGrid(dl=0.1)
+
+    See Also
+    --------
+
+    :class:`AutoGrid`
+        Specification for non-uniform grid along a given dimension.
+
+    **Notebooks:**
+        * `Photonic crystal waveguide polarization filter <../../../notebooks/PhotonicCrystalWaveguidePolarizationFilter.html>`_
+        * `Using automatic nonuniform meshing <../../notebooks/AutoGrid.html>`_
     """
 
     dl: pd.PositiveFloat = pd.Field(
@@ -284,6 +294,22 @@ class AutoGrid(GridSpec1d):
     Example
     -------
     >>> grid_1d = AutoGrid(min_steps_per_wvl=16, max_scale=1.4)
+
+    See Also
+    --------
+
+    :class:`UniformGrid`
+        Uniform 1D grid.
+
+    :class:`GridSpec`
+        Collective grid specification for all three dimensions.
+
+    **Notebooks:**
+        * `Using automatic nonuniform meshing <../../notebooks/AutoGrid.html>`_
+
+    **Lectures:**
+        *  `Time step size and CFL condition in FDTD <https://www.flexcompute.com/fdtd101/Lecture-7-Time-step-size-and-CFL-condition-in-FDTD/>`_
+        *  `Numerical dispersion in FDTD <https://www.flexcompute.com/fdtd101/Lecture-8-Numerical-dispersion-in-FDTD/>`_
     """
 
     min_steps_per_wvl: float = pd.Field(
@@ -316,7 +342,7 @@ class AutoGrid(GridSpec1d):
         description="The type of mesher to use to generate the grid automatically.",
     )
 
-    def _make_coords_initial(  # pylint:disable=too-many-arguments,arguments-differ,too-many-locals
+    def _make_coords_initial(
         self,
         axis: Axis,
         structures: List[StructureType],
@@ -412,24 +438,43 @@ class GridSpec(Tidy3dBaseModel):
     >>> custom = CustomGrid(dl=[0.2, 0.2, 0.1, 0.1, 0.1, 0.2, 0.2])
     >>> auto = AutoGrid(min_steps_per_wvl=12)
     >>> grid_spec = GridSpec(grid_x=uniform, grid_y=custom, grid_z=auto, wavelength=1.5)
+
+    See Also
+    --------
+
+    :class:`UniformGrid`
+        Uniform 1D grid.
+
+    :class:`AutoGrid`
+        Specification for non-uniform grid along a given dimension.
+
+    **Notebooks:**
+        * `Using automatic nonuniform meshing <../../notebooks/AutoGrid.html>`_
+
+    **Lectures:**
+        *  `Time step size and CFL condition in FDTD <https://www.flexcompute.com/fdtd101/Lecture-7-Time-step-size-and-CFL-condition-in-FDTD/>`_
+        *  `Numerical dispersion in FDTD <https://www.flexcompute.com/fdtd101/Lecture-8-Numerical-dispersion-in-FDTD/>`_
     """
 
     grid_x: GridType = pd.Field(
         AutoGrid(),
         title="Grid specification along x-axis",
         description="Grid specification along x-axis",
+        discriminator=TYPE_TAG_STR,
     )
 
     grid_y: GridType = pd.Field(
         AutoGrid(),
         title="Grid specification along y-axis",
         description="Grid specification along y-axis",
+        discriminator=TYPE_TAG_STR,
     )
 
     grid_z: GridType = pd.Field(
         AutoGrid(),
         title="Grid specification along z-axis",
         description="Grid specification along z-axis",
+        discriminator=TYPE_TAG_STR,
     )
 
     wavelength: float = pd.Field(
@@ -510,7 +555,6 @@ class GridSpec(Tidy3dBaseModel):
                     override_used[dl_axis] = True
         return override_used
 
-    # pylint:disable=too-many-locals,too-many-arguments
     def make_grid(
         self,
         structures: List[Structure],
@@ -556,7 +600,8 @@ class GridSpec(Tidy3dBaseModel):
                 log.warning(
                     f"Override structures take no effect along {axis_ind}-axis. "
                     "If intending to apply override structures to this axis, "
-                    "use 'AutoGrid'."
+                    "use 'AutoGrid'.",
+                    capture=False,
                 )
 
         grids_1d = [self.grid_x, self.grid_y, self.grid_z]
@@ -575,7 +620,7 @@ class GridSpec(Tidy3dBaseModel):
         return Grid(boundaries=coords)
 
     @classmethod
-    def auto(  # pylint:disable=too-many-arguments
+    def auto(
         cls,
         wavelength: pd.PositiveFloat = None,
         min_steps_per_wvl: pd.PositiveFloat = 10.0,
