@@ -1,4 +1,3 @@
-# pylint:disable=too-many-lines
 """Collection of functions for automatically generating a nonuniform grid. """
 
 from abc import ABC, abstractmethod
@@ -7,7 +6,7 @@ from math import isclose
 from itertools import compress
 import warnings
 
-import pydantic as pd
+import pydantic.v1 as pd
 import numpy as np
 from pyroots import Brentq
 from shapely.strtree import STRtree
@@ -17,18 +16,24 @@ from shapely.errors import ShapelyDeprecationWarning
 from ..base import Tidy3dBaseModel
 from ..types import Axis, ArrayFloat1D
 from ..structure import Structure, MeshOverrideStructure, StructureType
-from ..medium import PECMedium, Medium2D
+from ..medium import AnisotropicMedium, Medium2D, PECMedium
 from ...exceptions import SetupError, ValidationError
 from ...constants import C_0, fp_eps
+from ...log import log
 
 _ROOTS_TOL = 1e-10
 
+# Shrink min_step a little so that if e.g. a structure has target dl = 0.1 and a width of 0.1,
+# a grid point will be added on both sides of the structure. Without this factor, the mesher
+# ``is_close`` check will deem that the second point is too close.
+MIN_STEP_SCALE = 0.9999
+
 
 class Mesher(Tidy3dBaseModel, ABC):
     """Abstract class for automatic meshing."""
 
     @abstractmethod
-    def parse_structures(  # pylint:disable=too-many-arguments
+    def parse_structures(
         self,
         axis: Axis,
         structures: List[StructureType],
@@ -53,7 +58,6 @@ class GradedMesher(Mesher):
     """Implements automatic nonuniform meshing with a set minimum steps per wavelength and
     a graded mesh expanding from higher- to lower-resolution regions."""
 
-    # pylint:disable=too-many-locals, too-many-arguments
     def parse_structures(
         self,
         axis: Axis,
@@ -124,7 +128,7 @@ class GradedMesher(Mesher):
             structures_ordered, wavelength, min_steps_per_wvl, dl_min, axis
         )
         # Smallest of the maximum steps
-        min_step = np.amin(structure_steps)
+        min_step = MIN_STEP_SCALE * np.amin(structure_steps)
 
         # If empty simulation, return
         if len(structures) == 1:
@@ -137,45 +141,79 @@ class GradedMesher(Mesher):
         tree = self.bounds_2d_tree(struct_bbox)
 
         intervals = {"coords": list(domain_bounds), "structs": [[]]}
-        # Iterate in reverse order as latter structures override earlier ones. To properly handle
-        # containment then we need to populate interval coordinates starting from the top.
-        # If a structure is found to be completely contained, the corresponding ``struct_bbox`` is
-        # set to ``None``.
-        for str_ind in range(len(structures_ordered) - 1, -1, -1):
-            # 3D and 2D bounding box of current structure
-            bbox = struct_bbox[str_ind]
-            if bbox is None:
-                # Structure has been removed because it is completely contained
-                continue
-            bbox_2d = shapely_box(bbox[0, 0], bbox[0, 1], bbox[1, 0], bbox[1, 1])
+        """ Build the ``intervals`` dictionary. ``intervals["coords"]`` gets populated based on the
+        bounding boxes of all structures in the list (some filtering is done to exclude points that
+        will be too close together compared to the absolute lower desired step). At every point, the
+        ``"structs"`` list has length one lower than the ``"coords"`` list, and each element is
+        another list of all structure indexes that are found in the interval formed by ``coords[i]``
+        and ``coords[i + 1]``. This only includes structures that have a physical presence in the
+        interval, i.e. it excludes structures that are completely covered by higher-up ones.
 
-            # List of structure indexes that may intersect the current structure in 2D
-            try:
-                query_inds = tree.query_items(bbox_2d)
-            except AttributeError:
-                query_inds = tree.query(bbox_2d)
+        To build this, we iterate in reverse order as latter structures override earlier ones.
+        We also handle containment in the following way (note - we work with bounding boxes only):
+         1. If a structure that is lower in the list than the current structure is found to be
+          completely contained in 3D, the corresponding ``struct_bbox`` is immediately set to
+          ``None`` and nothing more will be done using that structure.
+         2. If the current structure is covering an interval but there's a higher-up structure that
+          contains it in 2D and also covers the same interval, then it will not be added to the
+          ``intervals["structs"] list for that interval.
+         3. If the current structure is found to not cover any interval, its bounding box
+          is set to ``None``, so that it will not affect structures that lie below it as per point
+          2. A warning is also raised since the structure will have an unpredictable effect on the
+          material coefficients used in the simulation.
+        """
 
-            # Remove all lower structures that the current structure completely contains
-            inds_lower = [
-                ind for ind in query_inds if ind < str_ind and struct_bbox[ind] is not None
-            ]
-            query_bbox = [struct_bbox[ind] for ind in inds_lower]
-            bbox_contains_inds = self.contains_3d(bbox, query_bbox)
-            for ind in bbox_contains_inds:
-                struct_bbox[inds_lower[ind]] = None
+        with log:
+            for str_ind in range(len(structures_ordered) - 1, -1, -1):
+                # 3D and 2D bounding box of current structure
+                bbox = struct_bbox[str_ind]
+                if bbox is None:
+                    # Structure has been removed because it is completely contained
+                    continue
+                bbox_2d = shapely_box(bbox[0, 0], bbox[0, 1], bbox[1, 0], bbox[1, 1])
 
-            # List of structure bboxes that contain the current structure in 2D
-            inds_upper = [ind for ind in query_inds if ind > str_ind]
-            query_bbox = [struct_bbox[ind] for ind in inds_upper if struct_bbox[ind] is not None]
-            bbox_contained_2d = self.contained_2d(bbox, query_bbox)
+                # List of structure indexes that may intersect the current structure in 2D
+                try:
+                    query_inds = tree.query_items(bbox_2d)
+                except AttributeError:
+                    query_inds = tree.query(bbox_2d)
 
-            # Handle insertion of the current structure bounds in the intervals
-            intervals = self.insert_bbox(intervals, str_ind, bbox, bbox_contained_2d, min_step)
+                # Remove all lower structures that the current structure completely contains
+                inds_lower = [
+                    ind for ind in query_inds if ind < str_ind and struct_bbox[ind] is not None
+                ]
+                query_bbox = [struct_bbox[ind] for ind in inds_lower]
+                bbox_contains_inds = self.contains_3d(bbox, query_bbox)
+                for ind in bbox_contains_inds:
+                    struct_bbox[inds_lower[ind]] = None
+
+                # List of structure bboxes that contain the current structure in 2D
+                inds_upper = [ind for ind in query_inds if ind > str_ind]
+                query_bbox = [
+                    struct_bbox[ind] for ind in inds_upper if struct_bbox[ind] is not None
+                ]
+                bbox_contained_2d = self.contained_2d(bbox, query_bbox)
+
+                # Handle insertion of the current structure bounds in the intervals
+                # The intervals list is modified in-place
+                too_small = self.insert_bbox(intervals, str_ind, bbox, bbox_contained_2d, min_step)
+                if too_small and (bbox[1, 2] - bbox[0, 2]) > 0:
+                    # If the structure is too small (but not 0D), issue a warning
+                    log.warning(
+                        f"A structure has a nonzero dimension along axis {'xyz'[axis]}, which "
+                        "is however too small compared to the generated mesh step along that "
+                        "direction. This could produce unpredictable results. We recommend "
+                        "increasing the resolution, or adding a mesh override structure to ensure "
+                        "that all geometries are at least one pixel thick along all dimensions."
+                    )
+                    # Also remove this structure from the bbox list so that lower structures
+                    # will not be affected by it, as it was not added to any interval.
+                    struct_bbox[str_ind] = None
 
         # Truncate intervals to domain bounds
         coords = np.array(intervals["coords"])
         num_ints = len(intervals["structs"])
-        in_domain = np.argwhere((coords >= domain_bounds[0]) * (coords <= domain_bounds[1]))
+        in_domain = np.nonzero((coords >= domain_bounds[0]) * (coords <= domain_bounds[1]))[0]
         intervals["coords"] = [intervals["coords"][int(i)] for i in in_domain]
         intervals["structs"] = [intervals["structs"][int(i)] for i in in_domain if i < num_ints]
 
@@ -197,7 +235,6 @@ class GradedMesher(Mesher):
 
         return np.array(intervals["coords"]), np.array(max_steps)
 
-    # pylint:disable=too-many-locals,too-many-arguments
     def insert_bbox(
         self,
         intervals: Dict[str, List],
@@ -229,17 +266,29 @@ class GradedMesher(Mesher):
             List of 3D bounding boxes that contain the current structure in 2D.
         min_step : float
             Absolute minimum interval size to impose.
+
+        Returns
+        -------
+        structure_too_small : bool
+            True if the structure did not span any interval after coordinates were inserted. This
+            would happen if the structure size is too small compared to the minimum step.
+
+        Note
+        ----
+            This function modifies ``intervals`` in-place.
         """
 
         coords = intervals["coords"]
         structs = intervals["structs"]
 
+        min_step_check = MIN_STEP_SCALE * min_step
+
         # Left structure bound
         bound_coord = str_bbox[0, 2]
-        indsmin = np.argwhere(bound_coord <= coords)
-        indmin = int(indsmin[0])  # coordinate is in interval index ``indmin - 1````
-        is_close_l = self.is_close(bound_coord, coords, indmin - 1, min_step)
-        is_close_r = self.is_close(bound_coord, coords, indmin, min_step)
+        indsmin = np.nonzero(bound_coord <= coords)[0]
+        indmin = int(indsmin[0])  # coordinate is in interval index ``indmin - 1``
+        is_close_l = self.is_close(bound_coord, coords, indmin - 1, min_step_check)
+        is_close_r = self.is_close(bound_coord, coords, indmin, min_step_check)
         is_contained = self.is_contained(bound_coord, bbox_contained_2d)
 
         # Decide on whether coordinate should be inserted or indmin modified
@@ -255,10 +304,10 @@ class GradedMesher(Mesher):
 
         # Right structure bound
         bound_coord = str_bbox[1, 2]
-        indsmax = np.argwhere(bound_coord >= coords)
+        indsmax = np.nonzero(bound_coord >= coords)[0]
         indmax = int(indsmax[-1])  # coordinate is in interval index ``indmax``
-        is_close_l = self.is_close(bound_coord, coords, indmax, min_step)
-        is_close_r = self.is_close(bound_coord, coords, indmax + 1, min_step)
+        is_close_l = self.is_close(bound_coord, coords, indmax, min_step_check)
+        is_close_r = self.is_close(bound_coord, coords, indmax + 1, min_step_check)
         is_contained = self.is_contained(bound_coord, bbox_contained_2d)
 
         # Decide on whether coordinate should be inserted or indmax modified
@@ -281,7 +330,7 @@ class GradedMesher(Mesher):
             if not self.is_contained(mid_coord, bbox_contained_2d):
                 structs[interval_ind].append(str_ind)
 
-        return {"coords": coords, "structs": structs}
+        return indmin >= indmax
 
     @staticmethod
     def reorder_structures_enforced_to_end(
@@ -376,13 +425,20 @@ class GradedMesher(Mesher):
         min_steps = []
         for structure in structures:
             if isinstance(structure, Structure):
-                if isinstance(structure.medium, (PECMedium, Medium2D)):
+                if isinstance(structure.medium, (PECMedium, Medium2D)) or (
+                    isinstance(structure.medium, AnisotropicMedium)
+                    and structure.medium.is_comp_pec(axis)
+                ):
+                    # for 2d medium, will always ignore even if not PEC;
+                    # later, this will be handled by _grid_corrections_2dmaterials
+                    # in simulation.py
                     index = 1.0
                 else:
                     n, k = structure.medium.eps_complex_to_nk(
-                        structure.medium.eps_diagonal(C_0 / wavelength)[axis]
+                        structure.medium.eps_diagonal(C_0 / wavelength)
                     )
-                    index = max(abs(n), abs(k))
+                    # take max among all directions because perpendicular eps defines wavelength
+                    index = max(max(abs(n)), max(abs(k)))
                 min_steps.append(max(dl_min, wavelength / index / min_steps_per_wvl))
             elif isinstance(structure, MeshOverrideStructure):
                 min_steps.append(max(dl_min, structure.dl[axis]))
@@ -390,7 +446,7 @@ class GradedMesher(Mesher):
 
     @staticmethod
     def rotate_structure_bounds(structures: List[StructureType], axis: Axis) -> List[ArrayFloat1D]:
-        """Get sturcture bounding boxes with a given ``axis`` rotated to z.
+        """Get structure bounding boxes with a given ``axis`` rotated to z.
 
         Parameters
         ----------
@@ -493,17 +549,17 @@ class GradedMesher(Mesher):
         # Re-compute minimum step in case some high-index structures were completely covered
         min_step = np.amin(max_steps)
 
-        # Filter interval coordintaes and max_steps
+        # Filter interval coordinates and max_steps
         coords_filter = [interval_coords[0]]
         steps_filter = []
         for coord_ind, coord in enumerate(interval_coords[1:]):
-            if coord - coords_filter[-1] > min_step:
+            if coord - coords_filter[-1] >= min_step:
                 coords_filter.append(coord)
                 steps_filter.append(max_steps[coord_ind])
 
         return coords_filter, steps_filter
 
-    def make_grid_multiple_intervals(  # pylint:disable=too-many-locals
+    def make_grid_multiple_intervals(
         self,
         max_dl_list: ArrayFloat1D,
         len_interval_list: ArrayFloat1D,
@@ -682,7 +738,6 @@ class GradedMesher(Mesher):
 
         return left_dl, right_dl
 
-    # pylint:disable=too-many-locals, too-many-return-statements, too-many-arguments
     def make_grid_in_interval(
         self,
         left_neighbor_dl: float,
@@ -754,7 +809,7 @@ class GradedMesher(Mesher):
             len_remaining = len_interval - len_scale
 
             # 1) interval length too small, cannot increase to large_dl, or barely can,
-            #    but the remaing part is less than large_dl
+            #    but the remaining part is less than large_dl
             if len_remaining < large_dl:
                 dl_list = self.grid_grow_in_interval(small_dl, max_scale, len_interval)
                 return dl_list if left_dl <= right_dl else np.flip(dl_list)
