@@ -1,18 +1,20 @@
-# pylint: disable=invalid-name
 """ utilities for plotting """
 from __future__ import annotations
 
 from typing import Any
 from functools import wraps
+import random
+import time
 
-import matplotlib.pylab as plt
-from matplotlib.patches import PathPatch
+import matplotlib.pyplot as plt
+from matplotlib.patches import PathPatch, ArrowStyle
 from matplotlib.path import Path
-from numpy import array, concatenate, ones
-import pydantic as pd
+from numpy import array, concatenate, ones, inf
+import pydantic.v1 as pd
 
 from .types import Ax
 from .base import Tidy3dBaseModel
+from ..exceptions import SetupError
 
 """ Constants """
 
@@ -24,28 +26,20 @@ ARROW_COLOR_SOURCE = "green"
 ARROW_COLOR_POLARIZATION = "brown"
 ARROW_ALPHA = 0.8
 
-
-# this times the min of axis height and width gives the arrow length
-ARROW_LENGTH_FACTOR = 0.1
-
-# this times ARROW_LENGTH gives width
-ARROW_WIDTH_FACTOR = 0.4
-
-# arrow width cannot be larger than this factor times the max of axis height and width
-MAX_ARROW_WIDTH_FACTOR = 0.02
-
+# Arrow length in inches
+ARROW_LENGTH = 0.3
 
 """ Decorators """
 
 
 def make_ax() -> Ax:
-    """makes an empty `ax`."""
+    """makes an empty ``ax``."""
     _, ax = plt.subplots(1, 1, tight_layout=True)
     return ax
 
 
 def add_ax_if_none(plot):
-    """Decorates `plot(*args, **kwargs, ax=None)` function.
+    """Decorates ``plot(*args, **kwargs, ax=None)`` function.
     if ax=None in the function call, creates an ax and feeds it to rest of function.
     """
 
@@ -87,6 +81,7 @@ class PlotParams(Tidy3dBaseModel):
     facecolor: Any = pd.Field(None, title="Face Color", alias="fc")
     fill: bool = pd.Field(True, title="Is Filled")
     hatch: str = pd.Field(None, title="Hatch Style")
+    zorder: float = pd.Field(None, title="Display Order")
     linewidth: pd.NonNegativeFloat = pd.Field(1, title="Line Width", alias="lw")
 
     def include_kwargs(self, **kwargs) -> PlotParams:
@@ -111,12 +106,16 @@ plot_params_geometry = PlotParams()
 plot_params_structure = PlotParams()
 plot_params_source = PlotParams(alpha=0.4, facecolor="limegreen", edgecolor="limegreen", lw=3)
 plot_params_monitor = PlotParams(alpha=0.4, facecolor="orange", edgecolor="orange", lw=3)
-plot_params_pml = PlotParams(alpha=0.7, facecolor="gray", edgecolor="gray", hatch="x")
-plot_params_pec = PlotParams(alpha=1.0, facecolor="gold", edgecolor="black")
-plot_params_pmc = PlotParams(alpha=1.0, facecolor="lightsteelblue", edgecolor="black")
-plot_params_bloch = PlotParams(alpha=1.0, facecolor="orchid", edgecolor="black")
-plot_params_symmetry = PlotParams(edgecolor="gray", facecolor="gray", alpha=0.6)
-plot_params_override_structures = PlotParams(linewidth=0.4, edgecolor="black", fill=False)
+plot_params_pml = PlotParams(alpha=0.7, facecolor="gray", edgecolor="gray", hatch="x", zorder=inf)
+plot_params_pec = PlotParams(alpha=1.0, facecolor="gold", edgecolor="black", zorder=inf)
+plot_params_pmc = PlotParams(alpha=1.0, facecolor="lightsteelblue", edgecolor="black", zorder=inf)
+plot_params_bloch = PlotParams(alpha=1.0, facecolor="orchid", edgecolor="black", zorder=inf)
+plot_params_symmetry = PlotParams(edgecolor="gray", facecolor="gray", alpha=0.6, zorder=inf)
+plot_params_override_structures = PlotParams(
+    linewidth=0.4, edgecolor="black", fill=False, zorder=inf
+)
+plot_params_fluid = PlotParams(facecolor="white", edgecolor="lightsteelblue", lw=0.4, hatch="xx")
+plot_params_grid = PlotParams(edgecolor="black", lw=0.2)
 
 # stores color of simulation.structures for given index in simulation.medium_map
 MEDIUM_CMAP = [
@@ -130,6 +129,13 @@ MEDIUM_CMAP = [
     "#877EBC",
 ]
 
+# colormap for structure's permittivity in plot_eps
+STRUCTURE_EPS_CMAP = "gist_yarg"
+STRUCTURE_HEAT_COND_CMAP = "gist_yarg"
+
+# default arrow style
+arrow_style = ArrowStyle.Simple(head_length=12, head_width=9, tail_width=4)
+
 
 """=================================================================================================
 Descartes modified from https://pypi.org/project/descartes/ for Shapely >= 1.8.0
@@ -206,8 +212,8 @@ def polygon_path(polygon):
 def polygon_patch(polygon, **kwargs):
     """Constructs a matplotlib patch from a geometric object
 
-    The `polygon` may be a Shapely or GeoJSON-like object with or without holes.
-    The `kwargs` are those supported by the matplotlib.patches.Polygon class
+    The ``polygon`` may be a Shapely or GeoJSON-like object with or without holes.
+    The ``kwargs`` are those supported by the matplotlib.patches.Polygon class
     constructor. Returns an instance of matplotlib.patches.PathPatch.
 
     Example
@@ -222,3 +228,42 @@ def polygon_patch(polygon, **kwargs):
 
 """End descartes modification
 ================================================================================================="""
+
+
+def plot_sim_3d(sim, width=800, height=800) -> None:
+    """Make 3D display of simulation in ipyython notebook."""
+
+    try:
+        from IPython.display import display, HTML
+    except ImportError as e:
+        raise SetupError(
+            "3D plotting requires ipython to be installed "
+            "and the code to be running on a jupyter notebook."
+        ) from e
+
+    uuid = str(int(time.time() * 1000)) + str(random.randint(0, 100000))
+
+    js_code = f"""
+    window.postMessageToViewer{uuid} = event => {{
+        if(event.data.type === 'viewer'&&event.data.uuid==='{uuid}'){{
+            document.getElementById('simulation-viewer{uuid}').contentWindow.postMessage({{ type: 'jupyter', uuid:'{uuid}', value:{sim._json_string}}}, '*')
+        }}
+    }};
+    window.addEventListener(
+        'message',
+        window.postMessageToViewer{uuid},
+        false
+    );
+    """
+    viewer_url = (
+        "https://feature-simulation-viewer.d3a9gfg7glllfq.amplifyapp.com/simulation-viewer?uuid="
+        + str(uuid)
+    )
+    html_code = f"""
+    <iframe id="simulation-viewer{uuid}" src={viewer_url} width="{width}" height="{height}" allowfullscreen="true"></iframe>
+    <script>
+        {js_code}
+    </script>
+    """
+
+    return display(HTML(html_code))
