@@ -5,7 +5,7 @@ import numpy as np
 import scipy.sparse as sp
 import scipy.sparse.linalg as spl
 
-from ...components.types import Numpy, ModeSolverType
+from ...components.types import Numpy, ModeSolverType, EpsSpecType
 from ...components.base import Tidy3dBaseModel
 from ...constants import ETA_0, C_0, fp_eps, pec_val
 from .derivatives import create_d_matrices as d_mats
@@ -18,6 +18,8 @@ TOL_COMPLEX = fp_eps
 TOL_EIGS = fp_eps
 # Tolerance for deciding on the matrix to be diagonal or tensorial
 TOL_TENSORIAL = 1e-6
+# shift target neff by this value, both rel and abs, whichever results in larger shift
+TARGET_SHIFT = 10 * fp_eps
 
 
 class EigSolver(Tidy3dBaseModel):
@@ -25,24 +27,25 @@ class EigSolver(Tidy3dBaseModel):
     It's a collection of static methods.
     """
 
-    # pylint:disable=too-many-statements,too-many-branches,too-many-locals
     @classmethod
-    def compute_modes(  # pylint:disable=too-many-arguments
+    def compute_modes(
         cls,
         eps_cross,
         coords,
         freq,
         mode_spec,
         symmetry=(0, 0),
-    ) -> Tuple[Numpy, Numpy]:
-        """Solve for the modes of a waveguide cross section.
+        direction="+",
+    ) -> Tuple[Numpy, Numpy, EpsSpecType]:
+        """
+        Solve for the modes of a waveguide cross-section.
 
         Parameters
         ----------
         eps_cross : array_like or tuple of array_like
             Either a single 2D array defining the relative permittivity in the cross-section,
-            or three 2D arrays defining the permittivity at the Ex, Ey, and Ez locations
-            of the Yee grid.
+            or nine 2D arrays defining the permittivity at the Ex, Ey, and Ez locations
+            of the Yee grid in the order xx, xy, xz, yx, yy, yz, zx, zy, zz.
         coords : List[Numpy]
             Two 1D arrays with each with size one larger than the corresponding axis of
             ``eps_cross``.
@@ -51,12 +54,15 @@ class EigSolver(Tidy3dBaseModel):
             (Hertz) Frequency at which the eigenmodes are computed.
         mode_spec : ModeSpec
             ``ModeSpec`` object containing specifications of the mode solver.
+        direction : Union["+", "-"]
+            Direction of mode propagation.
 
         Returns
         -------
-        Tuple[Numpy, Numpy]
-            The first array gives the E and H fields for all modes, the second one give sthe complex
-            effective index.
+        Tuple[Numpy, Numpy, str]
+            The first array gives the E and H fields for all modes, the second one gives the complex
+            effective index. The last variable describes permittivity characterization on the mode
+            solver's plane ("diagonal", "tensorial_real", or "tensorial_complex").
         """
 
         num_modes = mode_spec.num_modes
@@ -68,9 +74,11 @@ class EigSolver(Tidy3dBaseModel):
         k0 = omega / C_0
 
         if isinstance(eps_cross, Numpy):
-            eps_xx, eps_yy, eps_zz = eps_cross
-        elif len(eps_cross) == 3:
-            eps_xx, eps_yy, eps_zz = [np.copy(e) for e in eps_cross]
+            eps_xx, eps_xy, eps_xz, eps_yx, eps_yy, eps_yz, eps_zx, eps_zy, eps_zz = eps_cross
+        elif len(eps_cross) == 9:
+            eps_xx, eps_xy, eps_xz, eps_yx, eps_yy, eps_yz, eps_zx, eps_zy, eps_zz = (
+                np.copy(e) for e in eps_cross
+            )
         else:
             raise ValueError("Wrong input to mode solver pemittivity!")
 
@@ -87,9 +95,12 @@ class EigSolver(Tidy3dBaseModel):
         (2N, 2N), and the full tensorial case, in which case it has shape (4N, 4N)."""
         eps_tensor = np.zeros((3, 3, N), dtype=np.complex128)
         mu_tensor = np.zeros((3, 3, N), dtype=np.complex128)
-        for dim, eps in enumerate([eps_xx, eps_yy, eps_zz]):
-            eps_tensor[dim, dim, :] = eps.ravel()
-            mu_tensor[dim, dim, :] = 1.0
+        for row, eps_row in enumerate(
+            [[eps_xx, eps_xy, eps_xz], [eps_yx, eps_yy, eps_yz], [eps_zx, eps_zy, eps_zz]]
+        ):
+            mu_tensor[row, row, :] = 1.0
+            for col, eps in enumerate(eps_row):
+                eps_tensor[row, col, :] = eps.ravel()
 
         # Get Jacobian of all coordinate transformations. Initialize as identity (same as mu so far)
         jac_e = np.real(np.copy(mu_tensor))
@@ -98,7 +109,7 @@ class EigSolver(Tidy3dBaseModel):
         if bend_radius is not None:
             new_coords, jac_e, jac_h = radial_transform(new_coords, bend_radius, bend_axis)
 
-        if angle_theta > 0:
+        if np.abs(angle_theta) > 0:
             new_coords, jac_e_tmp, jac_h_tmp = angled_transform(new_coords, angle_theta, angle_phi)
             jac_e = np.einsum("ij...,jp...->ip...", jac_e_tmp, jac_e)
             jac_h = np.einsum("ij...,jp...->ip...", jac_h_tmp, jac_h)
@@ -108,8 +119,8 @@ class EigSolver(Tidy3dBaseModel):
         different from just the transformation of the derivative operator. For example, in a bent
         waveguide, there is strictly speaking no k-vector in the original coordinates as the system
         is not translationally invariant there. However, if we define kz = R k_phi, then the
-        effective index approaches that for a straight-waveguide in the limit of infinite radius. 
-        Since we use w = R phi in the radial_transform, there is nothing else neede in the k transform.
+        effective index approaches that for a straight-waveguide in the limit of infinite radius.
+        Since we use w = R phi in the radial_transform, there is nothing else needed in the k transform.
         For the angled_transform, the transformation between k-vectors follows from writing the field as
         E' exp(i k_p w) in transformed coordinates, and identifying this with
         E exp(i k_x x + i k_y y + i k_z z) in the original ones."""
@@ -157,10 +168,16 @@ class EigSolver(Tidy3dBaseModel):
             target = n_max
         else:
             target = mode_spec.target_neff
-        target_neff_p = target / np.linalg.norm(kp_to_k) + fp_eps
+        target_neff_p = target / np.linalg.norm(kp_to_k)
+
+        # shift target_neff slightly to avoid cases where the shiftted matrix is exactly singular
+        if abs(TARGET_SHIFT) > abs(target_neff_p * TARGET_SHIFT):
+            target_neff_p += TARGET_SHIFT
+        else:
+            target_neff_p *= 1 + TARGET_SHIFT
 
         # Solve for the modes
-        E, H, neff, keff = cls.solver_em(
+        E, H, neff, keff, eps_spec = cls.solver_em(
             Nx,
             Ny,
             eps_tensor,
@@ -169,39 +186,38 @@ class EigSolver(Tidy3dBaseModel):
             num_modes,
             target_neff_p,
             mode_spec.precision,
+            direction,
         )
 
-        # Filter polarization if needed
-        if mode_spec.filter_pol is not None:
-            te_int = np.sum(np.abs(E[0]) ** 2, axis=0) / np.sum(np.abs(E[:2]) ** 2, axis=(0, 1))
-            if mode_spec.filter_pol == "te":
-                sort_inds = np.concatenate(
-                    (np.nonzero(te_int >= 0.5)[0], np.nonzero(te_int < 0.5)[0])
-                )
-            elif mode_spec.filter_pol == "tm":
-                sort_inds = np.concatenate(
-                    (np.nonzero(te_int <= 0.5)[0], np.nonzero(te_int > 0.5)[0])
-                )
-            E = E[..., sort_inds]
-            H = H[..., sort_inds]
-            neff = neff[..., sort_inds]
-            keff = keff[..., sort_inds]
-
         # Transform back to original axes, E = J^T E'
         E = np.sum(jac_e[..., None] * E[:, None, ...], axis=0)
         E = E.reshape((3, Nx, Ny, 1, num_modes))
         H = np.sum(jac_h[..., None] * H[:, None, ...], axis=0)
         H = H.reshape((3, Nx, Ny, 1, num_modes))
         neff = neff * np.linalg.norm(kp_to_k)
+        keff = keff * np.linalg.norm(kp_to_k)
 
         fields = np.stack((E, H), axis=0)
 
-        return fields, neff + 1j * keff
+        if mode_spec.precision == "single":
+            # Recast to single precision which may have changed due to earlier manipulations
+            fields = fields.astype(np.complex64)
+
+        return fields, neff + 1j * keff, eps_spec
 
     @classmethod
     def solver_em(
-        cls, Nx, Ny, eps_tensor, mu_tensor, der_mats, num_modes, neff_guess, mat_precision
-    ):  # pylint:disable=too-many-arguments
+        cls,
+        Nx,
+        Ny,
+        eps_tensor,
+        mu_tensor,
+        der_mats,
+        num_modes,
+        neff_guess,
+        mat_precision,
+        direction,
+    ):
         """Solve for the electromagnetic modes of a system defined by in-plane permittivity and
         permeability and assuming translational invariance in the normal direction.
 
@@ -223,6 +239,8 @@ class EigSolver(Tidy3dBaseModel):
             Initial guess for the effective index.
         mat_precision : Union['single', 'double']
             Single or double-point precision in eigensolver.
+        direction : Union["+", "-"]
+            Direction of mode propagation.
 
         Returns
         -------
@@ -234,17 +252,65 @@ class EigSolver(Tidy3dBaseModel):
             Real part of the effective index, shape (num_modes, ).
         keff : np.ndarray
             Imaginary part of the effective index, shape (num_modes, ).
+        eps_spec : Union["diagonal", "tensorial_real", "tensorial_complex"]
+            Permittivity characterization on the mode solver's plane.
         """
 
+        # use a high-conductivity model for locations associated with a PEC
+        def conductivity_model_for_pec(eps, threshold=0.9 * pec_val):
+            """PEC entries associated with 'eps' are converted to a high-conductivity model."""
+            eps = eps.astype(complex)
+            eps[eps <= threshold] = 1 + 1j * np.abs(pec_val)
+            return eps
+
+        eps_tensor = conductivity_model_for_pec(eps_tensor)
+
         # Determine if ``eps`` and ``mu`` are diagonal or tensorial
         off_diagonals = (np.ones((3, 3)) - np.eye(3)).astype(bool)
         eps_offd = np.abs(eps_tensor[off_diagonals])
         mu_offd = np.abs(mu_tensor[off_diagonals])
-        is_tensorial = False
-        if np.any(eps_offd > TOL_TENSORIAL) or np.any(mu_offd > TOL_TENSORIAL):
-            is_tensorial = True
+        is_tensorial = np.any(eps_offd > TOL_TENSORIAL) or np.any(mu_offd > TOL_TENSORIAL)
 
-        # Determine data types
+        # initial vector for eigensolver in correct data type
+        vec_init = cls.set_initial_vec(Nx, Ny, is_tensorial=is_tensorial)
+
+        # call solver
+        kwargs = {
+            "eps": eps_tensor,
+            "mu": mu_tensor,
+            "der_mats": der_mats,
+            "num_modes": num_modes,
+            "neff_guess": neff_guess,
+            "vec_init": vec_init,
+            "mat_precision": mat_precision,
+        }
+
+        is_eps_complex = cls.isinstance_complex(eps_tensor)
+
+        if not is_tensorial:
+            eps_spec = "diagonal"
+            E, H, neff, keff = cls.solver_diagonal(**kwargs)
+            if direction == "-":
+                H[0] *= -1
+                H[1] *= -1
+                E[2] *= -1
+
+        elif not is_eps_complex:
+            eps_spec = "tensorial_real"
+            E, H, neff, keff = cls.solver_tensorial(**kwargs, direction="+")
+            if direction == "-":
+                E = np.conj(E)
+                H = -np.conj(H)
+
+        else:
+            eps_spec = "tensorial_complex"
+            E, H, neff, keff = cls.solver_tensorial(**kwargs, direction=direction)
+
+        return E, H, neff, keff, eps_spec
+
+    @classmethod
+    def matrix_data_type(cls, eps, mu, der_mats, mat_precision, is_tensorial):
+        """Determine data type that should be used for the matrix for diagonalization."""
         mat_dtype = np.float32
         # In tensorial case, even though the matrix can be real, the
         # expected eigenvalue is purely imaginary. So for now we enforce
@@ -253,13 +319,11 @@ class EigSolver(Tidy3dBaseModel):
             mat_dtype = np.complex128 if mat_precision == "double" else np.complex64
         else:
             # 1) check if complex or not
-            complex_solver = False
-            if (
-                cls.isinstance_complex(eps_tensor)
-                or cls.isinstance_complex(mu_tensor)
+            complex_solver = (
+                cls.isinstance_complex(eps)
+                or cls.isinstance_complex(mu)
                 or np.any([cls.isinstance_complex(f) for f in der_mats])
-            ):
-                complex_solver = True
+            )
             # 2) determine precision
             if complex_solver:
                 mat_dtype = np.complex128 if mat_precision == "double" else np.complex64
@@ -267,29 +331,36 @@ class EigSolver(Tidy3dBaseModel):
                 if mat_precision == "double":
                     mat_dtype = np.float64
 
-        # data type conversion
-        eps_tensor = cls.type_conversion(eps_tensor, mat_dtype)
-        mu_tensor = cls.type_conversion(mu_tensor, mat_dtype)
-        der_mats = [cls.type_conversion(f, mat_dtype) for f in der_mats]
-        neff_guess = cls.type_conversion(np.array([neff_guess]), mat_dtype)[0]
-
-        # initial vector for eigensolver
-        vec_init = cls.set_initial_vec(Nx, Ny, mat_dtype=mat_dtype, is_tensorial=is_tensorial)
-
-        # call solver
-        if is_tensorial:
-            return cls.solver_tensorial(
-                eps_tensor, mu_tensor, der_mats, num_modes, vec_init, neff_guess
-            )
-
-        return cls.solver_diagonal(eps_tensor, mu_tensor, der_mats, num_modes, vec_init, neff_guess)
+        return mat_dtype
 
     @classmethod
-    def solver_diagonal(
-        cls, eps, mu, der_mats, num_modes, vec_init, neff_guess
-    ):  # pylint:disable=too-many-arguments
+    def trim_small_values(cls, mat: sp.csr_matrix, tol: float) -> sp.csr_matrix:
+        """Eliminate elements of matrix ``mat`` for which ``abs(element) / abs(max_element) < tol``,
+        or ``np.abs(mat_data) < tol``. This operates in-place on mat so there is no return.
+        """
+        max_element = np.amax(np.abs(mat))
+        mat.data *= np.logical_or(np.abs(mat.data) / max_element > tol, np.abs(mat.data) > tol)
+        mat.eliminate_zeros()
+
+    @classmethod
+    def solver_diagonal(cls, eps, mu, der_mats, num_modes, neff_guess, vec_init, mat_precision):
         """EM eigenmode solver assuming ``eps`` and ``mu`` are diagonal everywhere."""
 
+        # code associated with these options is included below in case it's useful in the future
+        enable_incidence_matrices = False
+        enable_preconditioner = False
+        analyze_conditioning = False
+
+        def incidence_matrix_for_pec(eps_vec, threshold=0.9 * np.abs(pec_val)):
+            """Incidence matrix indicating non-PEC entries associated with 'eps_vec'."""
+            nnz = eps_vec[np.abs(eps_vec) < threshold]
+            eps_nz = eps_vec.copy()
+            eps_nz[np.abs(eps_vec) >= threshold] = 0
+            rows = np.arange(0, len(nnz))
+            cols = np.argwhere(eps_nz).flatten()
+            dnz = sp.csr_matrix(([1] * len(nnz), (rows, cols)), shape=(len(rows), len(eps_vec)))
+            return dnz
+
         mode_solver_type = "diagonal"
         N = eps.shape[-1]
 
@@ -302,9 +373,22 @@ class EigSolver(Tidy3dBaseModel):
         mu_zz = mu[2, 2, :]
         dxf, dxb, dyf, dyb = der_mats
 
+        def any_pec(eps_vec, threshold=0.9 * np.abs(pec_val)):
+            """Check if there are any PEC values in the given permittivity array."""
+            return np.any(np.abs(eps_vec) >= threshold)
+
+        if any(any_pec(i) for i in [eps_xx, eps_yy, eps_zz]):
+            enable_preconditioner = True
+
         # Compute the matrix for diagonalization
         inv_eps_zz = sp.spdiags(1 / eps_zz, [0], N, N)
         inv_mu_zz = sp.spdiags(1 / mu_zz, [0], N, N)
+
+        if enable_incidence_matrices:
+            dnz_xx, dnz_yy, dnz_zz = (incidence_matrix_for_pec(i) for i in [eps_xx, eps_yy, eps_zz])
+            dnz = sp.block_diag((dnz_xx, dnz_yy), format="csr")
+            inv_eps_zz = (dnz_zz.T) * dnz_zz * inv_eps_zz * (dnz_zz.T) * dnz_zz
+
         p11 = -dxf.dot(inv_eps_zz).dot(dyb)
         p12 = dxf.dot(inv_eps_zz).dot(dxb) + sp.spdiags(mu_yy, [0], N, N)
         p21 = -dyf.dot(inv_eps_zz).dot(dyb) - sp.spdiags(mu_xx, [0], N, N)
@@ -318,14 +402,53 @@ class EigSolver(Tidy3dBaseModel):
         qmat = sp.bmat([[q11, q12], [q21, q22]])
         mat = pmat.dot(qmat)
 
+        # Cast matrix to target data type
+        mat_dtype = cls.matrix_data_type(eps, mu, der_mats, mat_precision, is_tensorial=False)
+        mat = cls.type_conversion(mat, mat_dtype)
+
+        # Trim small values in single precision case
+        if mat_precision == "single":
+            cls.trim_small_values(mat, tol=fp_eps)
+
+        # Casting starting vector to target data type
+        vec_init = cls.type_conversion(vec_init, mat_dtype)
+
+        # Starting eigenvalue guess in target data type
+        eig_guess = cls.type_conversion(np.array([-(neff_guess**2)]), mat_dtype)[0]
+
+        if enable_incidence_matrices:
+            mat = dnz * mat * dnz.T
+            vec_init = dnz * vec_init
+
+        if enable_preconditioner:
+            precon = sp.diags(1 / mat.diagonal())
+            mat = mat * precon
+        else:
+            precon = None
+
+        if analyze_conditioning:
+            aca = mat.conjugate().T * mat
+            aac = mat * mat.conjugate().T
+            diff = aca - aac
+            print(spl.norm(diff, ord=np.inf), spl.norm(aca, ord=np.inf), spl.norm(aac, ord=np.inf))
+            print(spl.norm(diff, ord="fro"), spl.norm(aca, ord="fro"), spl.norm(aac, ord="fro"))
+
         # Call the eigensolver. The eigenvalues are -(neff + 1j * keff)**2
         vals, vecs = cls.solver_eigs(
             mat,
             num_modes,
             vec_init,
-            guess_value=-(neff_guess**2),
+            guess_value=eig_guess,
             mode_solver_type=mode_solver_type,
+            M=precon,
         )
+
+        if enable_preconditioner:
+            vecs = precon * vecs
+
+        if enable_incidence_matrices:
+            vecs = dnz.T * vecs
+
         neff, keff = cls.eigs_to_effective_index(vals, mode_solver_type)
 
         # Sort by descending neff
@@ -342,8 +465,8 @@ class EigSolver(Tidy3dBaseModel):
         h_field = qmat.dot(vecs)
         Hx = h_field[:N, :] / (1j * neff - keff)
         Hy = h_field[N:, :] / (1j * neff - keff)
-        Hz = inv_mu_zz.dot((dxf.dot(Ey) - dyf.dot(Ex)))
-        Ez = inv_eps_zz.dot((dxb.dot(Hy) - dyb.dot(Hx)))
+        Hz = inv_mu_zz.dot(dxf.dot(Ey) - dyf.dot(Ex))
+        Ez = inv_eps_zz.dot(dxb.dot(Hy) - dyb.dot(Hx))
 
         # Bundle up
         E = np.stack((Ex, Ey, Ez), axis=0)
@@ -356,8 +479,8 @@ class EigSolver(Tidy3dBaseModel):
 
     @classmethod
     def solver_tensorial(
-        cls, eps, mu, der_mats, num_modes, vec_init, neff_guess
-    ):  # pylint:disable=too-many-arguments
+        cls, eps, mu, der_mats, num_modes, neff_guess, vec_init, mat_precision, direction
+    ):
         """EM eigenmode solver assuming ``eps`` or ``mu`` have off-diagonal elements."""
 
         mode_solver_type = "tensorial"
@@ -425,13 +548,34 @@ class EigSolver(Tidy3dBaseModel):
             ]
         )
 
-        # Call the eigensolver. The eigenvalues are 1j * (neff + 1j * keff)
-        # Now multiply the matrix by -1j, so that eigenvalues are (neff + 1j * keff)
+        # The eigenvalues for the matrix above are 1j * (neff + 1j * keff)
+        # Multiply the matrix by -1j, so that eigenvalues are (neff + 1j * keff)
+        mat *= -1j
+
+        # change matrix sign for backward direction
+        if direction == "-":
+            mat *= -1
+
+        # Cast matrix to target data type
+        mat_dtype = cls.matrix_data_type(eps, mu, der_mats, mat_precision, is_tensorial=True)
+        mat = cls.type_conversion(mat, mat_dtype)
+
+        # Trim small values in single precision case
+        if mat_precision == "single":
+            cls.trim_small_values(mat, tol=fp_eps)
+
+        # Casting starting vector to target data type
+        vec_init = cls.type_conversion(vec_init, mat_dtype)
+
+        # Starting eigenvalue guess in target data type
+        eig_guess = cls.type_conversion(np.array([neff_guess]), mat_dtype)[0]
+
+        # Call the eigensolver.
         vals, vecs = cls.solver_eigs(
-            -1j * mat,
+            mat,
             num_modes,
             vec_init,
-            guess_value=neff_guess,
+            guess_value=eig_guess,
             mode_solver_type=mode_solver_type,
         )
         neff, keff = cls.eigs_to_effective_index(vals, mode_solver_type)
@@ -464,9 +608,7 @@ class EigSolver(Tidy3dBaseModel):
         return E, H, neff, keff
 
     @classmethod
-    def solver_eigs(
-        cls, mat, num_modes, vec_init, guess_value=1.0, **kwargs
-    ):  # pylint:disable=unused-argument
+    def solver_eigs(cls, mat, num_modes, vec_init, guess_value=1.0, M=None, **kwargs):
         """Find ``num_modes`` eigenmodes of ``mat`` cloest to ``guess_value``.
 
         Parameters
@@ -478,7 +620,9 @@ class EigSolver(Tidy3dBaseModel):
         guess_value : float, optional
         """
 
-        values, vectors = spl.eigs(mat, k=num_modes, sigma=guess_value, tol=TOL_EIGS, v0=vec_init)
+        values, vectors = spl.eigs(
+            mat, k=num_modes, sigma=guess_value, tol=TOL_EIGS, v0=vec_init, M=M
+        )
         return values, vectors
 
     @classmethod
@@ -525,12 +669,12 @@ class EigSolver(Tidy3dBaseModel):
         raise RuntimeError("Unsupported new_type.")
 
     @classmethod
-    def set_initial_vec(cls, Nx, Ny, mat_dtype=np.complex128, is_tensorial=False):
+    def set_initial_vec(cls, Nx, Ny, is_tensorial=False):
         """Set initial vector for eigs:
         1) The field at x=0 and y=0 boundaries are set to 0. This should be
         the case for PEC boundaries, but wouldn't hurt for non-PEC boundary;
         2) The vector is np.complex128 by default, and will be converted to
-        approriate type afterwards.
+        appropriate type afterwards.
 
         Parameters
         ----------
@@ -538,8 +682,6 @@ class EigSolver(Tidy3dBaseModel):
             Number of grids along x-direction.
         Ny : int
             Number of grids along y-direction.
-        mat_dtype : Union[np.complex128, np.complex64, np.float64, np.float32]
-            The type of returned vector.
         is_tensorial : bool
             diagonal or tensorial eigenvalue problem.
         """
@@ -553,7 +695,6 @@ class EigSolver(Tidy3dBaseModel):
         size = (Nx, Ny, len_multiplier)
         rng = np.random.default_rng(0)
         vec_init = rng.random(size) + 1j * rng.random(size)
-        vec_init = cls.type_conversion(vec_init, mat_dtype)
 
         # Set values at the boundary to be 0
         if Nx > 1:
@@ -599,6 +740,6 @@ class EigSolver(Tidy3dBaseModel):
         raise RuntimeError(f"Unidentified 'mode_solver_type={mode_solver_type}'.")
 
 
-def compute_modes(*args, **kwargs) -> Tuple[Numpy, Numpy]:
+def compute_modes(*args, **kwargs) -> Tuple[Numpy, Numpy, str]:
     """A wrapper around ``EigSolver.compute_modes``, which is used in ``ModeSolver``."""
     return EigSolver.compute_modes(*args, **kwargs)
