@@ -4,14 +4,16 @@ from __future__ import annotations
 from typing import Dict, Callable
 import json
 import functools
+
 import yaml
+import pydantic.v1 as pd
 
-import pydantic as pd
-
+from .log import log
 from .version import __version__
-from .log import FileError, SetupError, log
+from .exceptions import FileError, SetupError
 from .components.base import Tidy3dBaseModel
 
+
 """Storing version numbers."""
 
 
@@ -90,11 +92,13 @@ class Updater(pd.BaseModel):
     def from_file(cls, fname: str) -> Updater:
         """Dictionary representing the simulation loaded from file."""
 
+        # TODO: fix this, it broke
         if ".hdf5" in fname:
-            sim_dict = Tidy3dBaseModel.hdf5_to_dict(fname=fname)
+            sim_dict = Tidy3dBaseModel.from_file(fname=fname).dict()
+
         else:
             # try:
-            with open(fname, "r", encoding="utf-8") as f:
+            with open(fname, encoding="utf-8") as f:
                 if ".json" in fname:
                     sim_dict = json.load(f)
                 elif ".yaml" in fname:
@@ -122,7 +126,7 @@ class Updater(pd.BaseModel):
         return Version.from_string(version_string)
 
     def get_update_function(self):
-        """Get the highest update verion <= self.version."""
+        """Get the highest update version <= self.version."""
         leq_versions = [v for v in UPDATE_MAP if v <= self.version]
         if not leq_versions:
             raise SetupError(f"An update version <= {self.version} not found in update map.")
@@ -200,9 +204,101 @@ def iterate_update_dict(update_dict: Dict, update_types: Dict[str, Callable]):
             iterate_update_dict(item, update_types)
 
 
+@updates_from_version("1.8")
+def update_1_8(sim_dict: dict) -> dict:
+    """Updates version 1.8."""
+
+    def fix_missing_scalar_field(mnt_dict: dict) -> dict:
+        for key, val in mnt_dict["field_dataset"].items():
+            if isinstance(val, str) and val == "XR.DATAARRAY":
+                mnt_dict["field_dataset"][key] = "ScalarFieldDataArray"
+        return mnt_dict
+
+    iterate_update_dict(
+        update_dict=sim_dict,
+        update_types={
+            "CustomFieldSource": fix_missing_scalar_field,
+        },
+    )
+    return sim_dict
+
+
+@updates_from_version("1.7")
+def update_1_7(sim_dict: dict) -> dict:
+    """Updates version 1.7."""
+
+    def fix_angle_info(mnt_dict: dict) -> dict:
+        mnt_dict["type"] = "FieldProjectionAngleMonitor"
+        mnt_dict.pop("fields")
+        mnt_dict.pop("medium")
+        mnt_dict["proj_distance"] = 1e6
+        return mnt_dict
+
+    def fix_cartesian_info(mnt_dict: dict) -> dict:
+        mnt_dict["type"] = "FieldProjectionCartesianMonitor"
+        mnt_dict.pop("fields")
+        mnt_dict.pop("medium")
+        dist = mnt_dict.pop("plane_distance")
+        mnt_dict["proj_distance"] = dist
+        axis = mnt_dict.pop("plane_axis")
+        mnt_dict["proj_axis"] = axis
+        return mnt_dict
+
+    def fix_kspace_info(mnt_dict: dict) -> dict:
+        mnt_dict["type"] = "FieldProjectionKSpaceMonitor"
+        mnt_dict.pop("fields")
+        mnt_dict.pop("medium")
+        mnt_dict["proj_distance"] = 1e6
+        axis = mnt_dict.pop("u_axis")
+        mnt_dict["proj_axis"] = axis
+        return mnt_dict
+
+    def fix_diffraction_info(mnt_dict: dict) -> dict:
+        mnt_dict.pop("medium", None)
+        mnt_dict.pop("orders_x", None)
+        mnt_dict.pop("orders_y", None)
+        return mnt_dict
+
+    def fix_bloch_vec(mnt_dict: dict) -> dict:
+        mnt_dict["bloch_vec"] = mnt_dict["bloch_vec"]["real"]
+        return mnt_dict
+
+    iterate_update_dict(
+        update_dict=sim_dict,
+        update_types={
+            "Near2FarAngleMonitor": fix_angle_info,
+            "Near2FarCartesianMonitor": fix_cartesian_info,
+            "Near2FarKSpaceMonitor": fix_kspace_info,
+            "DiffractionMonitor": fix_diffraction_info,
+            "BlochBoundary": fix_bloch_vec,
+        },
+    )
+    return sim_dict
+
+
+@updates_from_version("1.6")
+def update_1_6(sim_dict: dict) -> dict:
+    """Updates version 1.6."""
+    if "grid_size" in sim_dict:
+        sim_dict.pop("grid_size")
+    return sim_dict
+
+
+@updates_from_version("1.5")
+def update_1_5(sim_dict: dict) -> dict:
+    """Updates version 1.5."""
+
+    def fix_mode_field_mnt(mnt_dict: dict) -> dict:
+        mnt_dict["type"] = "ModeSolverMonitor"
+        return mnt_dict
+
+    iterate_update_dict(update_dict=sim_dict, update_types={"ModeFieldMonitor": fix_mode_field_mnt})
+    return sim_dict
+
+
 @updates_from_version("1.4")
 def update_1_4(sim_dict: dict) -> dict:
-    """Updates version 1.3 to 1.4."""
+    """Updates version 1.4."""
 
     def fix_polyslab(geo_dict):
         """Fix a PolySlab dictionary."""
@@ -212,7 +308,7 @@ def update_1_4(sim_dict: dict) -> dict:
     def fix_modespec(ms_dict):
         """Fix a ModeSpec dictionary."""
         sort_by = ms_dict.pop("sort_by", None)
-        if sort_by != "largest_neff":
+        if sort_by and sort_by != "largest_neff":
             log.warning(
                 "ModeSpec.sort_by was removed in Tidy3D 1.5.0, reverting to sorting by "
                 "largest effective index. Use ModeSpec.filter_pol to select polarization instead."
@@ -235,7 +331,7 @@ def update_1_4(sim_dict: dict) -> dict:
 
 @updates_from_version("1.3")
 def update_1_3(sim_dict: dict) -> dict:
-    """Updates version 1.3 to 1.4."""
+    """Updates version 1.3."""
 
     sim_dict["boundary_spec"] = {"x": {}, "y": {}, "z": {}}
     for dim, pml_layer in zip(["x", "y", "z"], sim_dict["pml_layers"]):
