@@ -3,14 +3,18 @@ Commandline interface for tidy3d.
 """
 import json
 import os.path
+import ssl
 
 import click
 import requests
 import toml
 
-from tidy3d.web.cli.constants import TIDY3D_DIR, CONFIG_FILE, CREDENTIAL_FILE
-from tidy3d.web.cli.migrate import migrate
-from tidy3d.web.config import DEFAULT_CONFIG
+from ..cli.constants import TIDY3D_DIR, CONFIG_FILE, CREDENTIAL_FILE
+from ..cli.migrate import migrate
+from ..core.constants import KEY_APIKEY, HEADER_APIKEY
+from ..core.environment import Env
+from ..cli.converter import converter_arg
+from .develop.index import develop
 
 if not os.path.exists(TIDY3D_DIR):
     os.mkdir(TIDY3D_DIR)
@@ -25,10 +29,10 @@ def get_description():
     """
 
     if os.path.exists(CONFIG_FILE):
-        with open(CONFIG_FILE, "r", encoding="utf-8") as f:
+        with open(CONFIG_FILE, encoding="utf-8") as f:
             content = f.read()
             config = toml.loads(content)
-            return config.get("apikey", "")
+            return config.get(KEY_APIKEY, "")
     return ""
 
 
@@ -43,6 +47,18 @@ def tidy3d_cli():
 @click.option("--apikey", prompt=False)
 def configure(apikey):
     """Click command to configure the api key.
+
+    Parameters
+    ----------
+    apikey : str
+        User input api key.
+    """
+    configure_fn(apikey)
+
+
+def configure_fn(apikey: str) -> None:
+    """Python function that tries to set configuration based on a provided API key.
+
     Parameters
     ----------
     apikey : str
@@ -60,11 +76,11 @@ def configure(apikey):
         requests.Request
             Enriched request.
         """
-        req.headers["simcloud-api-key"] = apikey
+        req.headers[HEADER_APIKEY] = apikey
         return req
 
     if os.path.exists(CREDENTIAL_FILE):
-        with open(CREDENTIAL_FILE, "r", encoding="utf-8") as fp:
+        with open(CREDENTIAL_FILE, encoding="utf-8") as fp:
             auth_json = json.load(fp)
         email = auth_json["email"]
         password = auth_json["password"]
@@ -78,12 +94,18 @@ def configure(apikey):
         message = f"Current API key: [{current_apikey}]\n" if current_apikey else ""
         apikey = click.prompt(f"{message}Please enter your api key", type=str)
 
-    resp = requests.get(f"{DEFAULT_CONFIG.web_api_endpoint}/apikey", auth=auth, verify=False)
+    try:
+        resp = requests.get(
+            f"{Env.current.web_api_endpoint}/apikey", auth=auth, verify=Env.current.ssl_verify
+        )
+    except (requests.exceptions.SSLError, ssl.SSLError):
+        resp = requests.get(f"{Env.current.web_api_endpoint}/apikey", auth=auth, verify=False)
+
     if resp.status_code == 200:
         click.echo("Configured successfully.")
         with open(CONFIG_FILE, "w+", encoding="utf-8") as config_file:
             toml_config = toml.loads(config_file.read())
-            toml_config.update({"apikey": apikey})
+            toml_config.update({KEY_APIKEY: apikey})
             config_file.write(toml.dumps(toml_config))
     else:
         click.echo("API key is invalid.")
@@ -95,5 +117,15 @@ def migration():
     migrate()
 
 
+@click.command()
+@click.argument("lsf_file", type=click.Path(exists=True))
+@click.argument("new_file")
+def convert(lsf_file, new_file):
+    """Click command to convert .lsf project into Tidy3D .py file"""
+    converter_arg(lsf_file, new_file)
+
+
 tidy3d_cli.add_command(configure)
 tidy3d_cli.add_command(migration)
+tidy3d_cli.add_command(convert)
+tidy3d_cli.add_command(develop)
